import atexit
import configparser
import errno
import logging
import os
import threading

import click
import paramiko

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "-c",
    "--config",
    default="./config/",
    show_default=True,
    type=click.Path(exists=True, resolve_path=True),
    help="Path to config directory on local and remote machines",
)
@click.option(
    "-r",
    "--remote-config",
    "remote_config",
    type=str,
    help="Path to config directory on remote machine",
)
@click.option("-u", "--username", type=str, help="Username for remote server")
@click.option("-h", "--host", type=str, help="Hostname of remote server")
@click.option("-p", "--port", type=int, help="SSH port of remote server")
@click.option(
    "-k",
    "--key",
    type=click.Path(exists=True, resolve_path=True),
    help="Path to SSH key file",
)
@click.option(
    "-s",
    "--schedules",
    nargs=2,
    type=(click.Path(exists=True, resolve_path=True), click.Path()),
    help="Path to schedules directory on local and remote machines",
)
@click.option(
    "-i",
    "--images",
    nargs=2,
    type=(click.Path(exists=True, resolve_path=True), click.Path()),
    help="Path to images directory on local and remote machines",
)
@click.option(
    "-l",
    "--logs",
    nargs=2,
    type=(click.Path(exists=True, resolve_path=True), click.Path()),
    help="Path to logs directory on local and remote machines",
)
@click.version_option()
def syncfiles_cli(
    config, remote_config, username, host, port, key, schedules, images, logs
):
    logger.info("Starting syncfiles")
    logger.debug(
        f"syncfiles({username}, {host}, {port}, {key}, {config}, {schedules}, {images}, {logs})"
    )

    # Get local and remote config directories
    local_config_dir = config

    if os.path.isfile(local_config_dir):
        logger.info("Local config directory is a file, getting parent directory")
        local_config_dir = os.path.abspath(os.path.dirname(local_config_dir))

    # Get mtime of syncfiles.cfg
    syncfiles_mtime = os.path.getmtime(local_config_dir + "/syncfiles.cfg")

    # Read syncfiles.cfg
    (
        uname,
        hostname,
        p,
        k,
        remote_config_dir,
        local_schedules_dir,
        remote_schedules_dir,
        local_images_dir,
        remote_images_dir,
        local_logs_dir,
        remote_logs_dir,
    ) = _read_syncfiles_cfg(local_config_dir + "/syncfiles.cfg")

    # Override with any arguments passed in
    if remote_config is not None:
        remote_config_dir = remote_config
    if username is not None:
        uname = username
    if host is not None:
        hostname = host
    if port is not None:
        p = port
    if key is not None:
        k = key
    if schedules is not None:
        local_schedules_dir, remote_schedules_dir = schedules
    if images is not None:
        local_images_dir, remote_images_dir = images
    if logs is not None:
        local_logs_dir, remote_logs_dir = logs

    # Initial config directory sync
    logger.info("Performing initial config directory sync")
    cfg_ssh, cfg_scp = _get_client(uname, hostname, p, k)
    _sync_directory(cfg_scp, local_config_dir, remote_config_dir, "send", ".cfg")

    # Thread event
    threads_event = threading.Event()

    # Create clients and threads
    logger.info("Creating clients and threads")
    sch_ssh, sch_scp = _get_client(uname, hostname, p, k)
    sch_thread = threading.Thread(
        target=_continuous_sync,
        args=(
            sch_scp,
            local_schedules_dir,
            remote_schedules_dir,
            "both",
            ".ecsv",
            threads_event,
        ),
        daemon=True,
        name="sch_thread",
    )

    img_ssh, img_scp = _get_client(uname, hostname, p, k)
    img_thread = threading.Thread(
        target=_continuous_sync,
        args=(
            img_scp,
            local_images_dir,
            remote_images_dir,
            "send",
            (".fts", ".fits", ".fit"),
            threads_event,
        ),
        daemon=True,
        name="img_thread",
    )

    log_ssh, log_scp = _get_client(uname, hostname, p, k)
    log_thread = threading.Thread(
        target=_continuous_sync,
        args=(
            log_scp,
            local_logs_dir,
            remote_logs_dir,
            "send",
            (".log", ".ecsv"),
            threads_event,
        ),
        daemon=True,
        name="log_thread",
    )

    # Register exit function
    atexit.register(_close_connection, (cfg_ssh, sch_ssh, img_ssh, log_ssh))

    # Start threads
    logger.info("Starting threads")
    sch_thread.start()
    img_thread.start()
    log_thread.start()

    logger.info("Starting loop")
    while True:
        if os.path.getmtime(local_config_dir + "/syncfiles.cfg") > syncfiles_mtime:
            logger.info("syncfiles.cfg has been modified")

            # Update mtime
            logger.info("Updating mtime")
            syncfiles_mtime = os.path.getmtime(local_config_dir + "/syncfiles.cfg")

            # Stop threads from starting next loop sync and wait for them to finish
            logger.info("Stopping threads")
            threads_event.set()
            sch_thread.join()
            img_thread.join()
            log_thread.join()
            threads_event.clear()

            # End old ssh connections
            logger.info("Closing old connections")
            _close_connection((cfg_ssh, sch_ssh, img_ssh, log_ssh))

            # Re-parse the syncfiles.cfg
            logger.info("Re-parsing syncfiles.cfg")
            (
                uname,
                hostname,
                p,
                k,
                remote_config_dir,
                local_schedules_dir,
                remote_schedules_dir,
                local_images_dir,
                remote_images_dir,
                local_logs_dir,
                remote_logs_dir,
            ) = _read_syncfiles_cfg(local_config_dir + "/syncfiles.cfg")

            # Sync config directory
            logger.info("Syncing config directory")
            cfg_ssh, cfg_scp = _get_client(uname, hostname, p, k)
            _sync_directory(
                cfg_scp, local_config_dir, remote_config_dir, "send", ".cfg"
            )

            # Create clients and threads
            logger.info("Creating clients and threads")
            sch_ssh, sch_scp = _get_client(uname, hostname, p, k)
            sch_thread = threading.Thread(
                target=_continuous_sync,
                args=(
                    sch_scp,
                    local_schedules_dir,
                    remote_schedules_dir,
                    "both",
                    ".ecsv",
                    threads_event,
                ),
                daemon=True,
                name="sch_thread",
            )

            img_ssh, img_scp = _get_client(uname, hostname, p, k)
            img_thread = threading.Thread(
                target=_continuous_sync,
                args=(
                    img_scp,
                    local_images_dir,
                    remote_images_dir,
                    "send",
                    (".fts", ".fits", ".fit"),
                    threads_event,
                ),
                daemon=True,
                name="img_thread",
            )

            log_ssh, log_scp = _get_client(uname, hostname, p, k)
            log_thread = threading.Thread(
                target=_continuous_sync,
                args=(
                    log_scp,
                    local_logs_dir,
                    remote_logs_dir,
                    "send",
                    (".log", ".ecsv"),
                    threads_event,
                ),
                daemon=True,
                name="log_thread",
            )

            # Register exit function
            atexit.register(_close_connection, (cfg_ssh, sch_ssh, img_ssh, log_ssh))

            # Start threads
            logger.info("Starting threads")
            sch_thread.start()
            img_thread.start()
            log_thread.start()

        else:
            logger.debug("syncfiles.cfg has not been modified")
            _sync_directory(
                cfg_scp, local_config_dir, remote_config_dir, "send", ".cfg"
            )  # Sync config directory
            time.sleep(1)  # Sleep for 1 second


def _read_syncfiles_cfg(syncfiles_cfg):
    cfg = configparser.ConfigParser()
    cfg.read(syncfiles_cfg)

    uname = cfg.get("DEFAULT", "username", fallback="telrun")
    hostname = cfg["DEFAULT"]["hostname"]
    p = cfg.getint("DEFAULT", "port", fallback=22)
    k = cfg.get("DEFAULT", "key", fallback="./config/id_rsa")

    remote_config_dir = cfg.get(
        "DEFAULT", "remote_config_dir", fallback="/home/telrun/config/"
    )
    local_schedules_dir = cfg.get(
        "DEFAULT", "local_schedules_dir", fallback="./schedules/"
    )
    remote_schedules_dir = cfg.get(
        "DEFAULT", "remote_schedules_dir", fallback="/home/telrun/schedules/"
    )
    local_images_dir = cfg.get("DEFAULT", "local_images_dir", fallback="./images/")
    remote_images_dir = cfg.get(
        "DEFAULT", "remote_images_dir", fallback="/home/telrun/images/"
    )
    local_logs_dir = cfg.get("DEFAULT", "local_logs_dir", fallback="./logs/")
    remote_logs_dir = cfg.get(
        "DEFAULT", "remote_logs_dir", fallback="/home/telrun/logs/"
    )

    return (
        uname,
        hostname,
        p,
        k,
        remote_config_dir,
        local_schedules_dir,
        remote_schedules_dir,
        local_images_dir,
        remote_images_dir,
        local_logs_dir,
        remote_logs_dir,
    )


def _get_client(username, host, port, key):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(host, port, username, key_filename=key)
    scp = ssh.open_sftp()
    return (ssh, scp)


def _continuous_sync(scp, local_dir, remote_dir, mode, ext, event):
    while not event.is_set():
        _sync_directory(scp, local_dir, remote_dir, mode, ext)
        time.sleep(1)


def _sync_directory(scp, local_dir, remote_dir, mode, ext):
    logger.debug(f"_sync_directory({local_dir}, {remote_dir}, {mode}, {ext})")

    if mode == "send":
        try:
            scp.stat(remote_dir)
        except IOError as e:
            if e.errno == errno.ENOENT:
                logger.info("Creating remote directory %s", remote_dir)
                scp.mkdir(remote_dir)
            else:
                raise e
        scp.chdir(remote_dir)
        if not os.path.isdir(local_dir):
            raise NotADirectoryError(f"{local_dir} is not a directory")
        for f in os.listdir(local_dir):
            if not f.endswith(ext):
                continue
            if f in scp.listdir():
                if scp.stat(f).st_mtime < os.path.getmtime(local_dir + f):
                    logger.info(
                        "Putting local path %s to remote path %s",
                        local_dir + f,
                        remote_dir + f,
                    )
                    scp.put(local_dir + f, f)
            else:
                logger.info(
                    "Putting local path %s to remote path %s",
                    local_dir + f,
                    remote_dir + f,
                )
                scp.put(local_dir + f, f)
    elif mode == "receive":
        scp.chdir(remote_dir)
        if not os.path.isdir(local_dir):
            logger.info("Creating local directory %s", local_dir)
            os.mkdir(local_dir)
        for f in scp.listdir():
            if not f.endswith(ext):
                continue
            if f in os.listdir(local_dir):
                if scp.stat(f).st_mtime < os.path.getmtime(local_dir + f):
                    logger.info(
                        "Getting remote path %s to local path %s",
                        remote_dir + f,
                        local_dir + f,
                    )
                    scp.get(f, local_dir + f)
            else:
                logger.info(
                    "Getting remote path %s to local path %s",
                    remote_dir + f,
                    local_dir + f,
                )
                scp.get(f, local_dir + f)
    elif mode == "both":
        scp.chdir(remote_dir)
        for f in scp.listdir():
            if not f.endswith(ext):
                continue
            if f in os.listdir(local_dir):
                if scp.stat(f).st_mtime < os.path.getmtime(local_dir + f):
                    logger.info(
                        "Getting remote path %s to local path %s",
                        remote_dir + f,
                        local_dir + f,
                    )
                    scp.get(f, local_dir + f)
            else:
                logger.info(
                    "Getting remote path %s to local path %s",
                    remote_dir + f,
                    local_dir + f,
                )
                scp.get(f, local_dir + f)
        for f in os.listdir(local_dir):
            if not f.endswith(ext):
                continue
            if f in scp.listdir():
                if scp.stat(f).st_mtime < os.path.getmtime(local_dir + f):
                    logger.info(
                        "Putting local path %s to remote path %s",
                        local_dir + f,
                        remote_dir + f,
                    )
                    scp.put(local_dir + f, f)
            else:
                logger.info(
                    "Putting local path %s to remote path %s",
                    local_dir + f,
                    remote_dir + f,
                )
                scp.put(local_dir + f, f)


def _close_connection(sshs):
    for s in sshs:
        s.close()


syncfiles = syncfiles_cli.callback



"""
For each filter in Maxim's current filter list, perform one or
more autofocus runs to measure the focus offset required for each filter.
"""

import logging
import os
import sys
import time

import relimport
from iotalib import config, config_observatory, logutil, observatory

# Loop through the filter list this many times
NUM_MEASUREMENTS_PER_FILTER = 3

# List of filter index positions to measure.
# If many filters are known to use essentially the same
# filter position, you can save time by skipping those filters
# and manually editing the config file and duplicating focus
# positions across those filters.
FILTER_INDICES_TO_MEASURE = [0, 1, 8]

# The exposure length to use for each filter.
# This list must contain the same number of elements as FILTER_INDICES_TO_MEASURE.
# It can be useful to specify longer exposure lengths for narrowband filters.
FILTER_EXPOSURE_LENGTHS = [10, 10, 10]

# Template used to generate config file.
FOCUS_OFFSETS_FILE_TEMPLATE = """
# Focus values for each filter.
# This config file is parsed as a Python file, so any valid Python
# syntax (if statements, string formatting, etc) is allowed.

# AUTOGENERATED BY %(script_name)s on %(timestamp)s

# This Python dictionary maps a filter index to a focus value.
# Focus positions can be expressed in absolute or relative units.
# The important part is that the *difference* between any two filter
# measurements is an accurate offset that can be used when moving
# from one filter to the other.
# Note that it may not be valid to mix best-focus values that were
# measured at significantly different times or temperatures since
# the overall focus of the instrument may have changed.

best_focus_value = {
%(focus_values)s
}
"""

# Optionally write values to this config file
CONFIG_FILENAME = "focus_offsets.cfg"


def main():
    logutil.setup_log("measure_filter_focus_offsets.log")

    config_observatory.read()

    observatory.setup_mount()
    observatory.setup_camera()
    observatory.setup_focuser()
    observatory.setup_autofocus()

    best_focus_results = {}  # Map from filter index to a list of best focus values

    filter_names = observatory.camera.get_filter_names()

    for loop_number in range(NUM_MEASUREMENTS_PER_FILTER):
        for filter_index_index in range(len(FILTER_INDICES_TO_MEASURE)):
            filter_index = FILTER_INDICES_TO_MEASURE[filter_index_index]
            exp_length_seconds = FILTER_EXPOSURE_LENGTHS[filter_index_index]

            logging.info(
                "*** Iteration %d of %d, filter %d of %d ***",
                loop_number + 1,
                NUM_MEASUREMENTS_PER_FILTER,
                filter_index_index + 1,
                len(FILTER_INDICES_TO_MEASURE),
            )

            logging.info(
                "Setting filter %d (%s)", filter_index, filter_names[filter_index]
            )
            observatory.camera.set_active_filter(filter_index)

            logging.info("Setting exposure to %s seconds", exp_length_seconds)
            observatory.autofocus.set_exposure_length(exp_length_seconds)

            logging.info("Determining best focus position")
            focus_result = observatory.autofocus.run_autofocus()

            logging.info("Best focus result: %s", focus_result)

            best_focus_values_for_filter = best_focus_results.get(filter_index, [])
            if focus_result is not None:
                best_focus_values_for_filter.append(focus_result)
            best_focus_results[filter_index] = best_focus_values_for_filter

    config_lines = ""

    relative_to_value = None
    for filter_index in FILTER_INDICES_TO_MEASURE:
        filter_name = filter_names[filter_index]
        values = best_focus_results[filter_index]
        average_value = None
        if len(values) > 0:
            average_value = int(round(sum(values) / len(values)))

        if average_value is None:
            average_value_str = "None"
        else:
            average_value_str = "%.2f" % average_value

        values_str = ", ".join(["%.2f" % x for x in values])

        line = "%s: %s, # Filter: %s, Measurements: (%s)" % (
            filter_index,
            average_value_str,
            filter_name,
            values_str,
        )

        config_lines += "    " + line + "\n"

    logging.info("Final results:\n%s", config_lines)

    focus_offsets_cfg_text = FOCUS_OFFSETS_FILE_TEMPLATE % dict(
        script_name=os.path.basename(__file__),
        timestamp=time.ctime(),
        focus_values=config_lines,
    )

    print(focus_offsets_cfg_text)

    config_file_path = config.get_config_path(CONFIG_FILENAME)

    while True:
        response = input("Overwrite %s with new values? (y/n) " % CONFIG_FILENAME)
        response = response.strip().lower()
        if response == "n":
            break
        elif response == "y":
            f = open(config_file_path, "w")
            f.write(focus_offsets_cfg_text)
            f.close()
            logging.info("Configuration written to %s", config_file_path)
            break
        else:
            print("Invalid response")


if __name__ == "__main__":
    main()



import logging
import platform

from .device import Device
from .observatory_exception import ObservatoryException

logger = logging.getLogger(__name__)


class ASCOMDevice(Device):
    def __init__(self, identifier, alpaca=False, device_type="Device", **kwargs):
        logger.debug(f"ASCOMDevice.__init__({identifier}, alpaca={alpaca}, {kwargs})")
        self._identifier = identifier
        self._device = None

        if alpaca:
            import alpaca

            self._device = getattr(
                getattr(
                    __import__("alpaca." + device_type.lower()), device_type.lower()
                ),
                device_type,
            )(self._identifier, **kwargs)
        elif platform.system() == "Windows":
            from win32com.client import Dispatch

            self._device = Dispatch(self._identifier)
        else:
            raise ObservatoryException("If you are not on Windows, you must use Alpaca")

    def Action(self, ActionName, *ActionParameters):
        logger.debug(f"ASCOMDevice.Action({ActionName}, {ActionParameters})")
        return self._device.Action(ActionName, *ActionParameters)

    def CommandBlind(self, Command, Raw):  # pragma: no cover
        """
        .. deprecated:: 0.1.1
            ASCOM is deprecating this method.
        """

        logger.debug(f"ASCOMDevice.CommandBlind({Command}, {Raw})")
        self._device.CommandBlind(Command, Raw)

    def CommandBool(self, Command, Raw):  # pragma: no cover
        """
        .. deprecated:: 0.1.1
            ASCOM is deprecating this method.
        """

        logger.debug(f"ASCOMDevice.CommandBool({Command}, {Raw})")
        return self._device.CommandBool(Command, Raw)

    def CommandString(self, Command, Raw):  # pragma: no cover
        """
        .. deprecated:: 0.1.1
            ASCOM is deprecating this method.
        """

        logger.debug(f"ASCOMDevice.CommandString({Command}, {Raw})")
        return self._device.CommandString(Command, Raw)

    @property
    def Connected(self):
        logger.debug(f"ASCOMDevice.Connected property")
        return self._device.Connected

    @Connected.setter
    def Connected(self, value):
        logger.debug(f"ASCOMDevice.Connected property set to {value}")
        self._device.Connected = value

    @property
    def Description(self):
        logger.debug(f"ASCOMDevice.Description property")
        return self._device.Description

    @property
    def DriverInfo(self):
        logger.debug(f"ASCOMDevice.DriverInfo property")
        return self._device.DriverInfo

    @property
    def DriverVersion(self):
        logger.debug(f"ASCOMDevice.DriverVersion property")
        return self._device.DriverVersion

    @property
    def InterfaceVersion(self):
        logger.debug(f"ASCOMDevice.InterfaceVersion property")
        return self._device.InterfaceVersion

    @property
    def Name(self):
        logger.debug(f"ASCOMDevice.Name property")
        return self._device.Name

    @property
    def SupportedActions(self):
        logger.debug(f"ASCOMDevice.SupportedActions property")
        return self._device.SupportedActions



import logging
import os

import astropy.io.fits as pyfits

from .wcs import WCS

logger = logging.getLogger(__name__)


class AstrometryNetWCS(WCS):
    def __init__(self):
        logger.debug(f"AstrometryNetWCS.__init__() called")

        # Avoid documentation build failure by importing here
        from astroquery.astrometry_net import AstrometryNet

        self._solver = AstrometryNet()

    def Solve(self, filepath, **kwargs):
        logger.debug(f"AstrometryNetWCS.Solve({filepath}, {kwargs}) called")

        try_again = True
        submission_id = None
        while try_again:
            try:
                if not submission_id:
                    wcs_header = self._solver.solve_from_image(
                        filepath, submission_id=submission_id, **kwargs
                    )
                else:
                    wcs_header = self._solver.monitor_submission(
                        submission_id, solve_timeout=kwargs.get("solve_timeout", 120)
                    )
            except TimeoutError as e:
                submission_id = e.args[1]
            else:
                try_again = False

        if wcs_header:
            with pyfits.open(filepath, mode="update") as hdul:
                hdul[0].header.update(wcs_header)
                return True
        else:
            return False



import logging
import os
import platform
import shutil

import click

from .syncfiles import _get_client, _read_syncfiles_cfg, _sync_directory

logger = logging.getLogger(__name__)


@click.command()
@click.argument(
    "path",
    type=click.Path(resolve_path=True),
    default="./pyscope-observatory/",
    required=False,
)
@click.version_option()
def init_telrun_dir_cli(path):
    """test"""
    logger.debug(f"init_telrun_dir(path={path})")

    logger.info("Initializing a telrun home directory at %s" % path)

    if os.path.exists(path):
        logger.error(f"Path {path} already exists")
        raise click.Abort()

    logger.info(f"Creating directory {path}")
    os.mkdir(path)

    logger.info(f"Creating subdirectories")
    os.mkdir(os.path.join(path, "config"))
    os.mkdir(os.path.join(path, "schedules"))
    os.mkdir(os.path.join(path, "images"))
    os.mkdir(os.path.join(path, "logs"))

    logger.info("Creating empty config files")
    shutil.copyfile(
        os.path.join(os.path.dirname(__file__), "../config/telrun.cfg"),
        os.path.join(path, "config/telrun.cfg"),
    )

    shutil.copyfile(
        os.path.join(os.path.dirname(__file__), "../config/observatory.cfg"),
        os.path.join(path, "config/observatory.cfg"),
    )

    logger.info("Copying default shortcut startup script")
    if platform.system() == "Windows":
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "start_telrun.bat"),
            os.path.join(path, "start_telrun.bat"),
        )
    else:
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "start_telrun"),
            os.path.join(path, "start_telrun"),
        )

    logger.info("Done")


@click.command()
@click.argument(
    "path",
    type=click.Path(resolve_path=True, exists=True),
    default="./pyscope-observatory/",
    required=False,
)
@click.version_option()
def init_remote_dir_cli(path):
    logger.debug(f"init_remote_dir(path={path})")

    logger.info("Initializing a remote telrun home directory from %s" % path)

    if not os.path.exists(os.path.join(path, "config/syncfiles.cfg")):
        logger.info("No syncfiles.cfg found, creating one" % path)

        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "../config/syncfiles.cfg"),
            os.path.join(path, "config/syncfiles.cfg"),
        )

        logger.info(
            "Opening text editor... Please complete all fields and save the file"
        )
        if platform.system() == "Windows":
            os.system(os.path.join(path, "config/syncfiles.cfg"))
        else:
            try:
                logger.info("Trying to open $EDITOR")
                os.system(
                    "%s %s"
                    % (os.getenv("EDITOR"), os.path.join(path, "config/syncfiles.cfg"))
                )
            except:
                logger.warning("Could not open $EDITOR, trying to open vi")
                os.system("vi " + os.path.join(path, "config/syncfiles.cfg"))

        input("Press enter to continue...")

    logger.info("Reading syncfiles.cfg")

    local_config_dir = os.path.join(path, "config")

    (
        uname,
        hostname,
        p,
        k,
        remote_config_dir,
        local_schedules_dir,
        remote_schedules_dir,
        local_images_dir,
        remote_images_dir,
        local_logs_dir,
        remote_logs_dir,
    ) = _read_syncfiles_cfg(os.path.join(local_config_dir, "syncfiles.cfg"))

    logger.info("Performing initial single sync")

    logger.info("Getting SSH and SCP clients")
    ssh, scp = _get_client(uname, hostname, p, k)

    logger.info("Syncing config directory")
    _sync_directory(scp, local_config_dir, remote_config_dir, "send", ".cfg")

    logger.info("Syncing schedules directory")
    _sync_directory(scp, local_schedules_dir, remote_schedules_dir, "both", ".ecsv")

    logger.info("Syncing images directory")
    _sync_directory(
        scp, local_images_dir, remote_images_dir, "send", (".fts", ".fits", ".fit")
    )

    logger.info("Syncing logs directory")
    _sync_directory(scp, local_logs_dir, remote_logs_dir, "send", (".log", ".ecsv"))

    logger.info("Closing SSH and SCP clients")
    ssh.close()

    logger.info("Copying default shortcut startup script")
    if platform.system() == "Windows":
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "start_syncfiles.bat"),
            os.path.join(path, "start_syncfiles.bat"),
        )
    else:
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "start_syncfiles"),
            os.path.join(path, "start_syncfiles"),
        )


init_telrun_dir = init_telrun_dir_cli.callback
init_remote_dir = init_remote_dir_cli.callback



from abc import ABC, abstractmethod

from ._docstring_inheritee import _DocstringInheritee


class Rotator(ABC, metaclass=_DocstringInheritee):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def Halt(self):
        pass

    @abstractmethod
    def Move(self, Position):
        pass

    @abstractmethod
    def MoveAbsolute(self, Position):
        pass

    @abstractmethod
    def MoveMechanical(self, Position):
        pass

    @abstractmethod
    def Sync(self, Position):
        pass

    @property
    @abstractmethod
    def CanReverse(self):
        pass

    @property
    @abstractmethod
    def IsMoving(self):
        pass

    @property
    @abstractmethod
    def MechanicalPosition(self):
        pass

    @property
    @abstractmethod
    def Position(self):
        pass

    @property
    @abstractmethod
    def Reverse(self):
        pass

    @Reverse.setter
    @abstractmethod
    def Reverse(self, value):
        pass

    @property
    @abstractmethod
    def StepSize(self):
        pass

    @property
    @abstractmethod
    def TargetPosition(self):
        pass



import platform

import pytest

from pyscope.observatory import PWIAutofocus


@pytest.mark.skipif(platform.node() != "TCC1-MACRO", reason="Only works on TCC1-MACRO")
def test_pwi_autofocus():
    a = PWIAutofocus()

    best_position = a.Run(5, 120)



import ephem
from iotalib import convert


def calcOffsetPx(x, y):
    scale = 0.625 / 60.0  # arcmin/pixel for VAO SBIG camera
    dRA_fiber = 0.87
    dDec_fiber = 30.58  # Fiber offset from CCD field center
    xc = 1536
    yc = 1024  # pixel coords of CCD field center

    dRA_ccd = (xc - x) * scale  # Correct to center of CDD and scale RA
    dDec_ccd = (yc - y) * scale  # Correct to center of CDD and scale Dec
    dRA = dRA_fiber + dRA_ccd  # Correct fiber offset RA
    dDec = dDec_fiber + dDec_ccd  # Correct fiber offset Dec

    return (dRA, dDec)


def makeObsStar(ra_j2000_hours, dec_j2000_degs):
    ra_j2000_hours = convert.from_dms(ra_j2000_hours)
    dec_j2000_degs = convert.from_dms(dec_j2000_degs)
    obs = ephem.Observer()
    obs.date = e.now()
    obs.lat, obs.lon = "41.662195", "-91.532210"  # Van Allan Hall, Iowa City, Iowa, USA
    star = ephem.FixedBody()
    star._ra = convert.hours_to_rads(ra_j2000_hours)
    star._dec = convert.degs_to_rads(dec_j2000_degs)
    star.compute(obs)
    return star


def westofMeridian(ra_j2000_hours, dec_j2000_degs):
    star = makeObsStar(ra_j2000_hours, dec_j2000_degs)

    if convert.degs_to_rads(180) > star.az > convert.degs_to_rads(0):
        return false
    else:
        return true



import logging

import click
import numpy as np
from astropy.io import fits

logger = logging.getLogger(__name__)


@click.command(
    epilog="""Check out the documentation at
                https://pyscope.readthedocs.io/ for more
                information."""
)
@click.option(
    "-m",
    "--mode",
    type=click.Choice(["0", "1"]),
    default="0",
    show_choices=True,
    show_default=True,
    help="Mode to use for averaging FITS files (0 = median, 1 = mean).",
)
@click.option(
    "-o",
    "--outfile",
    type=click.Path(),
    required=True,
    help="Path to save averaged FITS file.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    show_default=True,
    help="Print verbose output.",
)
@click.argument("fnames", nargs=-1, type=click.Path(exists=True, resolve_path=True))
@click.version_option()
def avg_fits_cli(mode, outfile, fnames, verbose):
    if verbose:
        logger.setLevel(logging.DEBUG)

    logger.debug(f"avg_fits(mode={mode}, outfile={outfile}, fnames={fnames})")

    logger.info("Loading FITS files...")
    images = np.array([fits.open(fname)[0].data for fname in fnames])
    images = images.astype(np.float32)
    logger.info(f"Loaded {len(images)} FITS files")

    logger.info("Averaging FITS files...")
    if mode == "0":
        logger.debug("Calculating median...")
        image_avg = np.median(images, axis=0)
    elif mode == "1":
        logger.debug("Calculating mean...")
        image_avg = np.mean(images, axis=0)

    logger.debug(f"Image mean: {np.mean(image_avg)}")
    logger.debug(f"Image median: {np.median(image_avg)}")

    image_avg = image_avg.astype(np.uint16)

    logger.info(f"Saving averaged FITS file to {outfile}")

    with fits.open(fnames[-1]) as hdul:
        hdr = hdul[0].header

    hdr.add_comment(f"Averaged {len(images)} FITS files using pyscope")
    hdr.add_comment(f"Average mode: {mode}")
    fits.writeto(outfile, image_avg, hdr, overwrite=True)

    logger.info("Done!")


avg_fits = avg_fits_cli.callback



import pytest


def test_sensordescription(winer_oc):
    assert winer_oc.SensorDescription("rain_rate") == "RAIN"


def test_timesincelastupdate(winer_oc):
    assert winer_oc.TimeSinceLastUpdate("rain_rate") is not None


def test_averageperiod(winer_oc):
    assert winer_oc.AveragePeriod is None
    winer_oc.AveragePeriod = 10  # Note for this driver this does nothing
    assert winer_oc.AveragePeriod is None


def test_cloudcover(winer_oc):
    winer_oc.CloudCover is None


def test_dewpoint(winer_oc):
    winer_oc.DewPoint is None


def test_humidity(winer_oc):
    winer_oc.Humidity is not None


def test_pressure(winer_oc):
    winer_oc.Pressure is not None


def test_rainrate(winer_oc):
    winer_oc.RainRate is not None


def test_skybrightness(winer_oc):
    winer_oc.SkyBrightness is not None


def test_skyquality(winer_oc):
    winer_oc.SkyQuality is None


def test_skytemperature(winer_oc):
    winer_oc.SkyTemperature is None


def test_starfwhm(winer_oc):
    winer_oc.StarFWHM is not None


def test_temperature(winer_oc):
    winer_oc.Temperature is not None


def test_winddirection(winer_oc):
    winer_oc.WindDirection is not None


def test_windgust(winer_oc):
    winer_oc.WindGust is None


def test_windspeed(winer_oc):
    winer_oc.WindSpeed is not None



from abc import ABC, abstractmethod

from ._docstring_inheritee import _DocstringInheritee


class Camera(ABC, metaclass=_DocstringInheritee):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def AbortExposure(self):
        pass

    @abstractmethod
    def PulseGuide(self, Direction, Duration):
        pass

    @abstractmethod
    def StartExposure(self, Duration, Light):
        pass

    @abstractmethod
    def StopExposure(self):
        pass

    @property
    @abstractmethod
    def BayerOffsetX(self):
        pass

    @property
    @abstractmethod
    def BayerOffsetY(self):
        pass

    @property
    @abstractmethod
    def BinX(self):
        pass

    @BinX.setter
    @abstractmethod
    def BinX(self, value):
        pass

    @property
    @abstractmethod
    def BinY(self):
        pass

    @BinY.setter
    @abstractmethod
    def BinY(self, value):
        pass

    @property
    @abstractmethod
    def CameraState(self):
        pass

    @property
    @abstractmethod
    def CameraXSize(self):
        pass

    @property
    @abstractmethod
    def CameraYSize(self):
        pass

    @property
    @abstractmethod
    def CanAbortExposure(self):
        pass

    @property
    @abstractmethod
    def CanAsymmetricBin(self):
        pass

    @property
    @abstractmethod
    def CanFastReadout(self):
        pass

    @property
    @abstractmethod
    def CanGetCoolerPower(self):
        pass

    @property
    @abstractmethod
    def CanPulseGuide(self):
        pass

    @property
    @abstractmethod
    def CanSetCCDTemperature(self):
        pass

    @property
    @abstractmethod
    def CanStopExposure(self):
        pass

    @property
    @abstractmethod
    def CCDTemperature(self):
        pass

    @property
    @abstractmethod
    def CoolerOn(self):
        pass

    @CoolerOn.setter
    @abstractmethod
    def CoolerOn(self, value):
        pass

    @property
    @abstractmethod
    def CoolerPower(self):
        pass

    @property
    @abstractmethod
    def ElectronsPerADU(self):
        pass

    @property
    @abstractmethod
    def ExposureMax(self):
        pass

    @property
    @abstractmethod
    def ExposureMin(self):
        pass

    @property
    @abstractmethod
    def ExposureResolution(self):
        pass

    @property
    @abstractmethod
    def FastReadout(self):
        pass

    @FastReadout.setter
    @abstractmethod
    def FastReadout(self, value):
        pass

    @property
    @abstractmethod
    def FullWellCapacity(self):
        pass

    @property
    @abstractmethod
    def Gain(self):
        pass

    @Gain.setter
    @abstractmethod
    def Gain(self, value):
        pass

    @property
    @abstractmethod
    def GainMax(self):
        pass

    @property
    @abstractmethod
    def GainMin(self):
        pass

    @property
    @abstractmethod
    def Gains(self):
        pass

    @property
    @abstractmethod
    def HasShutter(self):
        pass

    @property
    @abstractmethod
    def HeatSinkTemperature(self):
        pass

    @property
    @abstractmethod
    def ImageArray(self):
        pass

    @property
    @abstractmethod
    def ImageReady(self):
        pass

    @property
    @abstractmethod
    def IsPulseGuiding(self):
        pass

    @property
    @abstractmethod
    def LastExposureDuration(self):
        pass

    @property
    @abstractmethod
    def LastExposureStartTime(self):
        pass

    @property
    @abstractmethod
    def MaxADU(self):
        pass

    @property
    @abstractmethod
    def MaxBinX(self):
        pass

    @property
    @abstractmethod
    def MaxBinY(self):
        pass

    @property
    @abstractmethod
    def NumX(self):
        pass

    @NumX.setter
    @abstractmethod
    def NumX(self, value):
        pass

    @property
    @abstractmethod
    def NumY(self):
        pass

    @NumY.setter
    @abstractmethod
    def NumY(self, value):
        pass

    @property
    @abstractmethod
    def Offset(self):
        pass

    @Offset.setter
    @abstractmethod
    def Offset(self, value):
        pass

    @property
    @abstractmethod
    def OffsetMax(self):
        pass

    @property
    @abstractmethod
    def OffsetMin(self):
        pass

    @property
    @abstractmethod
    def Offsets(self):
        pass

    @property
    @abstractmethod
    def PercentCompleted(self):
        pass

    @property
    @abstractmethod
    def PixelSizeX(self):
        pass

    @property
    @abstractmethod
    def PixelSizeY(self):
        pass

    @property
    @abstractmethod
    def ReadoutMode(self):
        pass

    @ReadoutMode.setter
    @abstractmethod
    def ReadoutMode(self, value):
        pass

    @property
    @abstractmethod
    def ReadoutModes(self):
        pass

    @property
    @abstractmethod
    def SensorName(self):
        pass

    @property
    @abstractmethod
    def SensorType(self):
        pass

    @property
    @abstractmethod
    def SetCCDTemperature(self):
        pass

    @SetCCDTemperature.setter
    @abstractmethod
    def SetCCDTemperature(self, value):
        pass

    @property
    @abstractmethod
    def StartX(self):
        pass

    @StartX.setter
    @abstractmethod
    def StartX(self, value):
        pass

    @property
    @abstractmethod
    def StartY(self):
        pass

    @StartY.setter
    @abstractmethod
    def StartY(self, value):
        pass

    @property
    @abstractmethod
    def SubExposureDuration(self):
        pass

    @SubExposureDuration.setter
    @abstractmethod
    def SubExposureDuration(self, value):
        pass



import photutils.background as photbackground
import photutils.segmentation as photsegmentation
from astropy import convolution, wcs
from astropy.io import fits
from photutils.utils import calc_total_error


def _get_image_source_catalog(
    image_path,
    box_size=50,
    filter_size=3,
    threshold=1.5,
    kernel_fwhm=3,
    kernel_size=5,
    effective_gain=1,
    connectivity=8,
    npixels=10,
    nlevels=32,
    contrast=0.001,
):
    """Finds sources in an image and returns a catalog of their positions
    along with other properties"""

    with fits.open(image_path) as hdul:
        image = hdul[0].data
        image = image.astype(float)
        hdr = hdul[0].header

    bkg = photbackground.Background2D(
        image,
        (box_size, box_size),
        filter_size=(filter_size, filter_size),
        bkg_estimator=photbackground.MedianBackground(),
    )
    image -= bkg.background

    kernel = photsegmentation.make_2dgaussian_kernel(kernel_fwhm, size=kernel_size)
    convolved_image = convolution.convolve(image, kernel)

    segment_map = photsegmentation.detect_sources(
        convolved_image,
        threshold * bkg.background_rms,
        npixels=npixels,
        connectivity=int(connectivity),
    )
    segm_deblend = photsegmentation.deblend_sources(
        convolved_image,
        segment_map,
        connectivity=int(connectivity),
        npixels=npixels,
        nlevels=nlevels,
        contrast=contrast,
        progress_bar=False,
    )

    err = calc_total_error(image, bkg.background_rms, effective_gain)

    cat = photsegmentation.SourceCatalog(
        image,
        segm_deblend,
        convolved_data=convolved_image,
        background=bkg.background,
        wcs=wcs.WCS(hdr),
        error=err,
    )

    return cat



import logging
import urllib.request

from ..utils import _get_number_from_line
from .safety_monitor import SafetyMonitor

logger = logging.getLogger(__name__)


class HTMLSafetyMonitor(SafetyMonitor):
    def __init__(self, url, check_phrase=b"ROOFPOSITION=OPEN"):
        logger.debug(
            f"""HTMLSafetyMonitor.__init__(
            {url}, check_phrase={check_phrase}) called"""
        )

        self._url = url
        self._check_phrase = check_phrase

    @property
    def IsSafe(self):
        logger.debug(f"""HTMLSafetyMonitor.IsSafe property called""")
        safe = None

        stream = urllib.request.urlopen(self._url)
        lines = stream.readlines()

        try:
            units = self._check_phrase.split(b" ")[1]
        except IndexError:
            units = ""

        for line in lines:
            s = _get_number_from_line(
                line,
                self._check_phrase.split(b"=")[0],
                units,
                False,
            )
            if s == self._check_phrase.split(b"=")[1]:
                safe = True
                break
        else:
            safe = False

        return safe



import logging

from .ascom_device import ASCOMDevice
from .cover_calibrator import CoverCalibrator

logger = logging.getLogger(__name__)


class ASCOMCoverCalibrator(ASCOMDevice, CoverCalibrator):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    def CalibratorOff(self):
        logger.debug(f"ASCOMCoverCalibrator.CalibratorOff() called")
        self._device.CalibratorOff()

    def CalibratorOn(self, Brightness):
        logger.debug(f"ASCOMCoverCalibrator.CalibratorOn({Brightness}) called")
        self._device.CalibratorOn(Brightness)

    def CloseCover(self):
        logger.debug(f"ASCOMCoverCalibrator.CloseCover() called")
        self._device.CloseCover()

    def HaltCover(self):
        logger.debug(f"ASCOMCoverCalibrator.HaltCover() called")
        self._device.HaltCover()

    def OpenCover(self):
        logger.debug(f"ASCOMCoverCalibrator.OpenCover() called")
        self._device.OpenCover()

    @property
    def Brightness(self):
        logger.debug(f"ASCOMCoverCalibrator.Brightness property called")
        return self._device.Brightness

    @property
    def CalibratorState(self):
        logger.debug(f"ASCOMCoverCalibrator.CalibratorState property called")
        return self._device.CalibratorState

    @property
    def CoverState(self):
        logger.debug(f"ASCOMCoverCalibrator.CoverState property called")
        return self._device.CoverState

    @property
    def MaxBrightness(self):
        logger.debug(f"ASCOMCoverCalibrator.MaxBrightness property called")
        return self._device.MaxBrightness



import pytest


def test_abortslew(device, disconnect):
    device.AbortSlew()


def test_closeshutter(device, disconnect):
    if device.CanSetShutter:
        device.CloseShutter()


def test_findhome(device, disconnect):
    if device.CanFindHome:
        device.FindHome()


def test_openshutter(device, disconnect):
    if device.CanSetShutter:
        device.OpenShutter()


def test_park(device, disconnect):
    if device.CanPark:
        device.Park()


def test_setpark(device, disconnect):
    if device.CanSetPark:
        device.SetPark()


def test_slewtoaltitude(device, disconnect):
    if device.CanSetAltitude:
        device.SlewToAltitude(45)


def test_sletoazimuth(device, disconnect):
    if device.CanSetAzimuth:
        device.SlewToAzimuth(45)


def test_synctoazimuth(device, disconnect):
    if device.CanSyncAzimuth:
        device.SyncToAzimuth(45)


def test_altitude(device, disconnect):
    assert device.Altitude is not None


def test_athome(device, disconnect):
    assert device.AtHome is not None


def test_atpark(device, disconnect):
    if device.CanPark:
        assert device.AtPark is not None


def test_azimuth(device, disconnect):
    assert device.Azimuth == 45


def test_shutterstatus(device, disconnect):
    assert device.ShutterStatus is not None


def test_slaved(device, disconnect):
    if device.CanSlave:
        assert device.Slaved is not None


def test_slewing(device, disconnect):
    assert device.Slewing is not None



#!/usr/bin/env python
"""
sexphot: Computes photometric magnitudes using sextractor, plots light curves
Optionally checks magnitudes using SDSS
N.B. Requires sextractor!

1.0 [command line sex, config file location defaults to /usr/local/sextractor/default.sex]
1.1 31 May 2016 - Fixed bug in Sloan lookup function
1.2 13 Jun 2016 - report median magnitudes on plots (if option -l)
1.3 21 Feb 2017 - skip stars with fluxerr = 0
1.4 31 May 2017 - Correctly [?] account for airmass, average color
2.0  2 Dec 2018 - add BJD, solve for minimum time, set ZP mags for IKON camera, changed max  FWHM to 5 pixs, remove SDSS
2.1 10 Jan 2019 - IF AIRMASS keyword not found, calculate using ELEVATION keyword (or give up, use 1.0), add fwhm_filter
2.11 2 Feb 2019 - Check: if sextractor couldn't find stars = skip (try/except)
2.2 17 May 2020 - Fix array integer index problem in time of minimum solver
2.3 22 Jan 2021 - add pDF plot format
2.4 11 Jan 2022 - check Maxim version: do not add 0.5x exposure time to time of observation if version = 6.22+
                  since DATE-OBS and JD keywords are now == observation midpoint (Maxim v.6.22+)
"""
vers = "2.4 (11 Jan 2021)"

import glob
import os
import re
import sys
import warnings
from optparse import OptionParser

import matplotlib.pyplot as plt
import numpy as np
from astropy import coordinates as coord
from astropy import time
from astropy import units as u
from astropy.coordinates import SkyCoord

# from astropy.time import Time
# from astropy import units as u
from astropy.io import fits
from astropy.io.fits import getheader, setval, update
from matplotlib.pyplot import cm
from scipy.optimize import curve_fit, minimize
from scipy.stats import chi2

# Avoid annoying warning about matplotlib building the font cache
warnings.filterwarnings("ignore")

# Sextractor config file path
sex_path = "/usr/local/sextractor/default.sex"


def get_args():
    global parser
    parser = OptionParser(
        description="%prog computes photometric magnitudes using sextractor, plots light curves",
        version="%s" % vers,
    )
    parser.add_option(
        "-s",
        dest="sigma",
        metavar="sigma",
        action="store",
        type=float,
        default=5,
        help="Sextractor detection threshold [default 5]",
    )
    parser.add_option(
        "-c",
        dest="config",
        metavar="config",
        action="store",
        help="phot config file name [no default]",
    )
    parser.add_option(
        "-d",
        dest="datafile",
        metavar="outfile",
        action="store",
        default="",
        help="Output csv file name",
    )
    parser.add_option(
        "-F",
        dest="fwhm_off",
        metavar="fwhm_off",
        action="store_true",
        default=False,
        help="Skip FWHM check, default = False",
    )
    parser.add_option(
        "-l",
        dest="line",
        metavar="line",
        action="store_true",
        default=False,
        help="Plot median line",
    )
    parser.add_option(
        "-j",
        dest="jdrange",
        metavar="jdrange",
        action="store",
        default="0,0",
        help="BJD range (BJDmin, BJDmax)",
    )
    parser.add_option(
        "-P",
        dest="PDF",
        metavar="PDF",
        action="store_true",
        default=False,
        help="PDF plot format[default png format]",
    )
    parser.add_option(
        "-t",
        dest="tmin",
        metavar="Time_minimum",
        action="store_true",
        default=False,
        help="Solve for time of minimum [default False]",
    )
    parser.add_option(
        "-v",
        dest="verbose",
        metavar="Verbose",
        action="store_true",
        default=False,
        help="Verbose output",
    )
    parser.add_option(
        "-w",
        dest="width",
        metavar="width",
        action="store",
        type=int,
        default=20,
        help="Minimum fit width, sample times [default 20]",
    )
    parser.add_option(
        "-y",
        dest="yrange",
        metavar="yrange",
        action="store",
        default="0,0",
        help="Differential plot yrange [default: autoscale]",
    )
    parser.add_option(
        "-z",
        dest="zp",
        metavar="Zeropoint",
        action="store",
        type=float,
        default=-1,
        help="Zero-point magnitude, defaults to FITS header value",
    )
    return parser.parse_args()


def hms2rad(hms_str):
    import re

    result = 0
    fields = re.split(r"[: _]", hms_str)
    fields = [float(x) for x in fields]
    while len(fields) > 0:
        result = result / 60.0 + fields.pop()
    rad = result * np.pi / 12.0
    return rad


def get_hdrdata(ftsfile):
    hdr = getheader(ftsfile, 0)
    date = hdr["DATE-OBS"]
    exptime = hdr["EXPTIME"]
    filter = hdr["FILTER"][0]
    if "AIRMASS" in hdr:
        airmass = hdr["AIRMASS"]
    elif "ELEVATION" in hdr:
        airmass = 1 / np.sin(hms2rad(hdr["ELEVATION"]))
    else:
        if verbose:
            print(
                "%s: No airmass or elevation in header, assuming airmass = 1.0"
                % ftsfile
            )
        airmass = 1.0
    zp = 0
    zperr = 0
    if "ZMAG" in hdr:
        zp = hdr["ZMAG"]
    if "ZMAGERR" in hdr:
        zperr = hdr["ZMAGERR"]
    nbin = hdr["XBINNING"]  # Assume same for y binning
    arcsec_pixel = np.abs(hdr["CDELT1"] * 3600.0)
    ra_str = hdr["RA"]
    dec_str = hdr["DEC"]
    Maxim_version = float(hdr["SWCREATE"].split()[3])
    jd_utc = hdr["JD"]
    if Maxim_version < 6.22:
        if verbose:
            print(
                "Maxim version %.2f (< 6.22), adding 0.5x exposure time to time of observation"
                % Maxim_version
            )
        jd_utc += exptime / (2.0 * 86400)
    bjd = calc_bjd(jd_utc, ra_str, dec_str)
    return (
        bjd,
        date,
        ra_str,
        dec_str,
        exptime,
        filter,
        arcsec_pixel,
        airmass,
        nbin,
        zp,
        zperr,
    )


def get_sexinfo(sexname, fwhm_filter, exptime, scale):
    fn = open(sexname, "r")
    lines = fn.readlines()[15:]
    Nr = []
    Ra = []
    Dec = []
    Snr = []
    Flux = []
    Fluxerr = []
    Fwhm = []
    V = []
    Verr = []
    for line in lines:
        (
            nr,
            dum,
            dum,
            flux,
            fluxerr,
            x_pix,
            y_pix,
            ra_deg,
            dec_deg,
            profile_x,
            profile_y,
            pa,
            fwhm_pixel,
            dum,
            flag,
        ) = [float(x) for x in line.split()]
        v = -2.5 * np.log10(flux / exptime)
        if fluxerr == 0:
            continue
        snr = flux / fluxerr
        verr = 2.5 * (fluxerr / flux)  # Expanding log10(1+x) ~ 2.5x
        Nr.append(nr)
        Ra.append(ra_deg)
        Dec.append(dec_deg)
        Flux.append(flux)
        Fluxerr.append(fluxerr)
        Fwhm.append(fwhm_pixel * np.abs(scale))
        Snr.append(snr)
        V.append(v)
        Verr.append(verr)
    fn.close()
    # Trim list to stars by restricting fwhm values
    fwhm_min = 1.4
    fwhm_max = 7.0
    A = list(zip(Ra, Dec, Snr, Flux, Fluxerr, Fwhm, V, Verr))
    B = []
    for j in range(len(A)):
        if fwhm_min < A[j][5] < fwhm_max or fwhm_off:
            B.append(A[j])
    Ra, Dec, Snr, Flux, Fluxerr, Fwhm, V, Verr = list(zip(*B))
    V = np.array(V)
    Verr = np.array(Verr)
    return Nr, Ra, Dec, Snr, Flux, Fluxerr, Fwhm, V, Verr


def phase_diff(jd, jd0, P):
    # Returns phase and o-c [day] at heliocentric observed date jd given ephemeris jd0,P
    jd_diff = jd - jd0
    phase = (jd_diff % P) / P
    if phase >= 0.5:
        o_c = (phase - 1) * P
    else:
        o_c = phase * P
    return phase, o_c


def fgauss(x, a, b, x0, w):
    t = (x - x0) ** 2 / w**2
    return a * np.exp(-t) + b


def trim(indices, A):
    # Trims arrays packed in A, dropping elements with given indices
    B = []
    for a in A:
        B.append(np.delete(a, indices))
    return B


def get_magnitudes(Ra, Dec, Nr_sex, Ra_sex, Dec_sex, max_diff, Mag_sex, Mag_sex_err):
    N1 = len(Ra)
    N2 = len(Ra_sex)
    Mag = np.empty(N1) * np.nan
    Mag_err = np.empty(N1) * np.nan
    Nr = np.empty(N1) * np.nan
    for j in range(N1):
        for k in range(N2):
            dra = np.abs(Ra[j] - Ra_sex[k])
            ddec = np.abs(Dec[j] - Dec_sex[k])
            if dra < max_diff and ddec < max_diff:
                Mag[j] = Mag_sex[k]
                Mag_err[j] = Mag_sex_err[k]
                Nr[j] = Nr_sex[k]
    Mag = np.array(Mag)
    Mag_err = np.array(Mag_err)
    return Nr, Mag, Mag_err


def parse_config(config_file):
    Objects = []
    Filters = []
    Ftsfiles = []
    Mag_catalog = []
    Ra_hms = []
    Dec_dms = []
    Ra_deg = []
    Dec_deg = []
    if not os.path.isfile(config_file):
        sys.exit("Configuration file %s does not exist, try again" % config_file)
    else:
        fn = open(config_file, "r")
        lines = fn.readlines()
        fn.close()
        for line in lines:
            line = line.split()
            if line == []:
                continue  # Skip blank lines
            elif line[0] == "I":
                Ftsfiles.append(line[1])
            elif line[0] == "F":
                Filter = line[1]
            elif line[0] == "S":
                object, ra_hms, dec_dms = line[1:4]
                Objects.append(object)
                Ra_hms.append(ra_hms)
                Dec_dms.append(dec_dms)
                c = SkyCoord(ra_hms, dec_dms, unit=(u.hourangle, u.deg), frame="icrs")
                Ra_deg.append(c.ra.deg)
                Dec_deg.append(c.dec.deg)
            elif line[0] == "T":
                title = " ".join(line[1:])
            elif line[0] == "E":
                BJD0, P0 = [float(s) for s in line[1:]]
    return (
        Objects,
        Filter,
        Ftsfiles,
        Ra_hms,
        Dec_dms,
        Ra_deg,
        Dec_deg,
        Mag_catalog,
        title,
        BJD0,
        P0,
    )


def calc_bjd(jd_utc, ra_str, dec_str):
    """
    Calculate barycentric dynamical Julian date from UTC Julian date, source coordinates
    Refs: http://docs.astropy.org/en/stable/time
    Eastman, et al. 2010 PASP,122,935
    """
    object = coord.SkyCoord(ra_str, dec_str, unit=(u.hourangle, u.deg), frame="icrs")
    lowell = coord.EarthLocation.of_site(
        "Multiple Mirror Telescope"
    )  # close enough to Winer
    times = time.Time(jd_utc, format="jd", scale="utc", location=lowell)
    ltt_bary = times.light_travel_time(object)
    bjd_tdb = times.tdb + ltt_bary
    return bjd_tdb.value


def calc_tmin(BJD, obj, obj_sigma, width):
    jd_frac, jd_int = np.modf(BJD)
    jd1 = jd_int[0]  # Integer part of first BJD time
    jmin = np.argmax(
        obj
    )  # index of maximum (minimum magnitude), use as guess for fitted minimum
    BJD_min0 = BJD[jmin]
    if verbose:
        print(
            "Found sample minimum at BJD = %.5f (jmin = %i, mag = %.2f)"
            % (BJD_min0, jmin, obj[jmin])
        )
    w = int(width / 2)  # Width of fit in points
    # print(jmin,w)
    x = np.array(BJD[jmin - w : jmin + w]) - BJD[0]
    y = obj[jmin - w : jmin + w]
    s = obj_sigma[jmin - w : jmin + w]

    # weighted Gaussian fit
    init_vals = (obj[jmin], 0, BJD[jmin] - BJD[0], width * exptime / 86400)
    [a_fit, b_fit, x0_fit, w_fit], cov = curve_fit(fgauss, x, y, p0=init_vals)
    [a_s, b_s, x0_s, w_s] = np.sqrt(np.diag(cov))
    delta_jd = x0_fit
    sigma = x0_s  # Uncertainty in delta_jd
    bjd_min = BJD[0] + delta_jd

    # generate a model for plotting
    npts = 100
    xmod = np.linspace(BJD[jmin - w] - BJD[0], BJD[jmin + w] - BJD[0], npts)
    ymod = fgauss(xmod, a_fit, b_fit, x0_fit, w_fit)
    xmod += BJD[0]
    # phs_mod = ( (xmod-jd0) % P)/P
    return bjd_min, sigma, xmod, ymod


# ======== MAIN ================

# Max difference: config vs Sex position [deg]
max_diff = 5 / 3600.0

# Define dictionary of zero-point values and extinction for filters [guesses except for Sloan G, R, I]
# Cal_Apogee = { 'N':(22.0,0.20), 'B':(21.5,0.35), 'G':(22.68,0.28), 'V':(20.6,0.20), 'R':(20.3,0.12), 'I':(20.66,0.05) }
Cal = {
    "N": (23.2, 0.20),
    "C": (23.2, 0.20),
    "L": (23.2, 0.20),
    "B": (22.3, 0.35),
    "G": (22.68, 0.28),
    "V": (22.5, 0.20),
    "R": (22.45, 0.12),
    "I": (21.90, 0.05),
}

# Get command  line arguments, assign parameter values
(opts, args) = get_args()


if not opts.config:
    parser.error("config file (-c) not given, try again")
config_file = opts.config
detect_threshold = opts.sigma  # Sextractor detection threshold [sigma]
csvfile = opts.datafile  # optional CSV output filename
fwhm_off = opts.fwhm_off  # FWHM filter [on by default]
jdmin, jdmax = [float(x) for x in opts.jdrange.split(",")]  # Julian date range
line = opts.line  # Plot median line
solve_tmin = opts.tmin  # Solve for minimum time
verbose = opts.verbose  # Print diagnostics, more
width = opts.width  # Width for finding minimum
ymin, ymax = [
    float(x) for x in opts.yrange.split(",")
]  # Differential plot width, magnitudes
zp_user = opts.zp  # Zeropoint magnitude
PDF = opts.PDF  # Use PDF plot format?

# Parse configuration file
(
    Objects,
    Filter,
    Ftsfiles,
    Ra_hms,
    Dec_dms,
    Ra_deg,
    Dec_deg,
    Mag_catalog,
    title,
    BJD0,
    P0,
) = parse_config(config_file)
nstar = len(Objects)

BJD = []
Date = []
Mag_all = []
Mag_err_all = []
FitsFile_all = []
Nr_all = []

# Expand filenames if needed
if "*" in Ftsfiles[0] or "?" in Ftsfiles[0]:
    Ftsfiles = glob.glob(Ftsfiles[0])
if verbose:
    print("Reading %i FITS image files" % len(Ftsfiles))
for ftsfile in Ftsfiles:
    # Get useful header info [NB not currently using nbin]
    try:
        (
            bjd,
            date,
            ra_str,
            dec_str,
            exptime,
            filter,
            scale,
            airmass,
            nbin,
            zp,
            zperr,
        ) = get_hdrdata(ftsfile)
    except:
        if verbose:
            print("%s header does not have required keywords, skipping" % ftsfile)
        continue
    # If wrong filter, skip
    if filter != Filter:
        if verbose:
            print(
                "%s: Wrong filter [expecting %s, got %s], skipping"
                % (ftsfile, Filter, filter)
            )
        continue

    # If not in user-specified JD range, skip
    if jdmin == 0 and jdmax == 0:
        pass
    else:
        if not jdmin <= bjd <= jdmax:
            if verbose:
                print(
                    "%s: BJD %.5f not in range %.5f - %.5f, skipping"
                    % (ftsfile, bjd, jdmin, jdmax)
                )
            continue
    # Run sextractor
    sexname = os.path.basename(ftsfile).split(".")[0] + ".sexout"
    if verbose:
        print(
            "Running sextractor on %s with detection threshold = %.1f sigma"
            % (ftsfile, detect_threshold)
        )
    os.system(
        "sex %s -c %s -CATALOG_NAME %s -DETECT_THRESH %.1f -VERBOSE_TYPE QUIET"
        % (ftsfile, sex_path, sexname, detect_threshold)
    )

    # Get position, magnitude info for each listed star in output file
    try:
        (
            Nr_sex,
            Ra_sex,
            Dec_sex,
            Snr,
            Flux,
            Fluxerr,
            Fwhm_sex,
            Mag_sex,
            Mag_sex_err,
        ) = get_sexinfo(sexname, fwhm_off, exptime, scale)
        nobs = len(Ra_sex)
        if verbose:
            print("Sextractor found %i stars" % nobs)
    except:
        if verbose:
            print("Sextractor could'nt find stars, skipping %s" % ftsfile)
        continue
    # Get magnitudes for target objects using position match to sextractor output
    Nr, Mag, Mag_err = get_magnitudes(
        Ra_deg, Dec_deg, Nr_sex, Ra_sex, Dec_sex, max_diff, Mag_sex, Mag_sex_err
    )
    # Convert to magnitude by adding ZP and correcting for extinction. Use user-supplied ZP if specified
    if zp_user > 0:
        ZP = zp_user
        if verbose:
            print("Using user-supplied zero-point (ZP = %.2f)" % ZP)
    elif zp > 0:
        ZP = zp
        if verbose:
            print("Using zero-point found in FITS header: %.2f)" % ZP)
    else:
        ZP = Cal[Filter][0]
        if verbose:
            print("Using preset zero-point for %s filter: (ZP = %.2f)" % (filter, ZP))
    k = Cal[Filter][1]
    # correct for airmass, assume average color correction 0.1
    Mag -= k * airmass - 0.1
    Mag += ZP

    # Add to array, but only if all stars detected
    if not np.isnan(Mag).any():
        BJD.append(bjd)
        Date.append(date)
        Mag_all.append(Mag)
        Mag_err_all.append(Mag_err)
        FitsFile_all.append(ftsfile)
        Nr_all.append(Nr)

nepoch = len(BJD)
if verbose:
    print("Analyzing %i images" % nepoch)

# Sort by BJD
BJD, FitsFile_all, Date, Nr_all, Mag_all, Mag_err_all = (
    list(x)
    for x in zip(*sorted(zip(BJD, FitsFile_all, Date, Nr_all, Mag_all, Mag_err_all)))
)

# Convert to numpy arrays
Mag = np.array(Mag_all)
Mag_err = np.array(Mag_err_all)

# Subtract reference star magnitudes
Ref_Mag = Mag[:, -1]
Diff_mags = Mag - Ref_Mag[:, np.newaxis]

# Target star is first listed
Mag_Target = Diff_mags[:, 0]
Mag_Target_Err = Mag_err[:, 0]
Name_Target = Objects[0].strip()

# Solve for minimum if requested
if solve_tmin:
    bjd_min, sigma, xmod, ymod = calc_tmin(BJD, Mag_Target, Mag_Target_Err, width)
    phs_min, oc = phase_diff(bjd_min, BJD0, P0)
    if verbose:
        print(
            "Fitted minimum at BJD: %.5f +/- %.5f, O-C = %.5f days (%.1f sec +/- %.1f sec). Phase at min = %.3f)"
            % (bjd_min, sigma, oc, oc * 86400, sigma * 86400, phs_min)
        )
else:
    bjd_min = 0
    phs_min = 0
    oc = 0
    jd1 = 0
    sigma = 0
    xmod = 0
    phs_mod = 0
    ymod = 0

# Calculate median differential magnitudes
Medians = np.median(Mag, axis=0)
Stds = np.std(Mag, axis=0)
"""
Diff_mag += Medians[-1]
Diff_err = np.sqrt(Mag_err**2 + Mag_err[-1]**2)
"""
# Plot light curves

# First plot all stars on same plot
fig = plt.figure(figsize=(12, 8))
for j in range(nepoch):
    mjd = BJD[j] - BJD[0]
    color = iter(cm.rainbow(np.linspace(0, 1, nstar + 1)))
    for k in range(nstar):
        c = next(color)
        if j == 0:
            plt.errorbar(
                mjd,
                Mag[j][k],
                yerr=Mag_err[j][k],
                marker="o",
                markersize=5,
                c=c,
                label="%s" % Objects[k],
            )
        else:
            plt.errorbar(
                mjd, Mag[j][k], yerr=Mag_err[j][k], marker="o", markersize=5, c=c
            )
plt.title(title)
plt.legend(loc=2, fontsize=8)
ymin0 = np.max(Medians) + 1
ymax0 = np.min(Medians) - 1
plt.ylim(ymin0, ymax0)
if jdmin != 0.0:
    plt.xlim(jdmin - BJD[0], jdmax - BJD[0])
plt.ylabel("%s Magnitude" % Filter)
plt.xlabel("Days since BJD %.5f (%s)" % (BJD[0], Date[0]))
plt.grid(True)
if PDF:
    plot_title = "%s_lc-all.pdf" % Name_Target
else:
    plot_title = "%s_lc-all.png" % Name_Target
plt.savefig(plot_title)
print("Saved light curve of all stars as %s" % plot_title)
plt.close(fig)

# Plot target star differential magnitude with minimum fit [optional]
fig = plt.figure(figsize=(12, 8))
BJD0_int = np.int(BJD[0])
mBJD = np.array(BJD) - BJD0_int

plt.errorbar(
    mBJD, Mag_Target, yerr=Mag_Target_Err, ls="none", marker="o", markersize=5, c="b"
)
plt.suptitle(title, fontsize=14)
if solve_tmin:
    plt.plot(xmod - BJD0_int, ymod, "r--")
    title = (
        "BJD0: %.6f, P0: %.10f,\n BJDmin: %.5f +/- %.5f,  O-C: %.5f days [%.1f s +/- %.1f s]"
        % (BJD0, P0, bjd_min, sigma, oc, oc * 86400, sigma * 86400)
    )
else:
    title = "%s [RA: %s, Dec: %s] Filter = %s" % (
        Name_Target,
        Ra_hms[0],
        Dec_dms[0],
        Filter,
    )
plt.title(title, fontsize=12)

# Set plot limits
if jdmin != 0.0:
    plt.xlim(jdmin - BJD0_int, jdmax - BJD0_int)
if ymin != 0 or ymax != 0:
    plt.ylim(ymin, ymax)
else:
    mag_range = np.abs(max(Mag_Target) - min(Mag_Target))
    ymin = min(Mag_Target) - 0.2 * mag_range
    ymax = max(Mag_Target) + 0.2 * mag_range
    plt.ylim(ymax, ymin)

UTDate, UTTime = Date[0].split("T")
plt.xlabel("Barycentric JD - %i (Start: %s, %s UT)" % (BJD0_int, UTDate, UTTime))
plt.ylabel("Differential Magnitude (Filter %s)" % Filter)
plt.grid(True)
if PDF:
    plot_file = "%s_%s_lc.pdf" % (Name_Target, UTDate)
else:
    plot_file = "%s_%s_lc.png" % (Name_Target, UTDate)
plt.savefig(plot_file)
print("Saved differential l.c. plot %s" % plot_file)
plt.close(fig)


# write CSV output file if requested
if csvfile != "":
    fn = open(csvfile, "w")
    for j in range(nepoch):
        str1 = " ".join(
            "%s  %.3f %.3f  " % (Objects[k], Mag[j][k], Mag_err[j][k])
            for k in range(nstar)
        )
        s = "%10.5f  %30s   %s\n" % (BJD[j], FitsFile_all[j], str1)
        fn.write(s)
    print("Wrote data file: %s" % csvfile)
    fn.close()



import logging

from .ascom_device import ASCOMDevice
from .switch import Switch

logger = logging.getLogger(__name__)


class ASCOMSwitch(ASCOMDevice, Switch):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    def CanWrite(self, ID):
        logger.debug(f"ASCOMSwitch.CanWrite({ID}) called")
        return self._device.CanWrite(ID)

    def GetSwitch(self, ID):
        logger.debug(f"ASCOMSwitch.GetSwitch({ID}) called")
        return self._device.GetSwitch(ID)

    def GetSwitchDescription(self, ID):
        logger.debug(f"ASCOMSwitch.GetSwitchDescription({ID}) called")
        return self._device.GetSwitchDescription(ID)

    def GetSwitchName(self, ID):
        logger.debug(f"ASCOMSwitch.GetSwitchName({ID}) called")
        return self._device.GetSwitchName(ID)

    def MaxSwitchValue(self, ID):
        logger.debug(f"ASCOMSwitch.MaxSwitchValue({ID}) called")
        return self._device.MaxSwitchValue(ID)

    def MinSwitchValue(self, ID):
        logger.debug(f"ASCOMSwitch.MinSwitchValue({ID}) called")
        return self._device.MinSwitchValue(ID)

    def SetSwitch(self, ID, State):
        logger.debug(f"ASCOMSwitch.SetSwitch({ID}, {State}) called")
        self._device.SetSwitch(ID, State)

    def SetSwitchName(self, ID, Name):
        logger.debug(f"ASCOMSwitch.SetSwitchName({ID}, {Name}) called")
        self._device.SetSwitchName(ID, Name)

    def SetSwitchValue(self, ID, Value):
        logger.debug(f"ASCOMSwitch.SetSwitchValue({ID}, {Value}) called")
        self._device.SetSwitchValue(ID, Value)

    def SwitchStep(self, ID):
        logger.debug(f"ASCOMSwitch.SwitchStep({ID}) called")
        self._device.SwitchStep(ID)

    @property
    def MaxSwitch(self):
        logger.debug("ASCOMSwitch.MaxSwitch property called")
        return self._device.MaxSwitch



import math
import os
import sys
import tempfile
import time
from datetime import datetime

import astropy.io.fits as fits
import ephem
import numpy
import pulsar_dimmer
import relimport
from iotalib import convert, rst, talonwcs
from win32com.client import Dispatch

"""
This script attempts to put a target RA and Dec at a particular pixel position
(which may be outside the bounds of the physical CCD in the case of an off-axis
spectrometer fiber).

The script slews to a target, takes an image, finds a WCS solution and offsets
the mount to place the target RA/Dec at the requested pixel position. This process
repeats until the target is positioned within a specified tolerance or the maximum
number of attempts has been exceeded.

The target pixel position can be calibrated by manually positioning a star at the
target position and running the calibrate_target_pixel script.
"""

### CONFIGURATION VALUES ########################

# configuration_file = sys.argv[1]
configuration_file = "calibration_config.txt"
config = []
with open(configuration_file) as infile:
    for line in infile:
        config.append((line.split()[0]))

# MOUNT_DRIVER = "ASCOM.SoftwareBisque.Telescope"  # Use this for the Paramount ME (controlled through TheSky)
MOUNT_DRIVER = "SiTech.Telescope"
HOME_MOUNT = True

OBJECT_NAME = ""  # Provides the name of an object to search for in the SIMBAD Database (If empty string, use provided coordinates)
TARGET_AZ = float(config[1])  # RA coordinates of the target star, in J2000 hours
TARGET_ALT = float(config[2])  # Dec coordinates of the target star, in J2000 degrees
SETTLE_TIME_SECONDS = float(
    config[3]
)  # Pause for this many seconds after slewing to each target before taking an image

BINNING = float(
    config[4]
)  # Image binning (higher binning reduces resolution but speeds up image readout time)

EXPOSURE_LENGTH_SECONDS_DARK = float(
    config[8]
)  # Exposure length of each dark image, in seconds
NUMBER_DARKS = float(config[6])
NUMBER_BIAS = float(config[7])
NUMBER_FLATS = float(config[5])
FILTER_NAMES = config[9].split(",")  # ["L","R","V","B","H","G","W","N"]
# FILTER_EXPOSURE_TIME = [2,2,2.5,7,15,0,2,2]
FILTER_EXPOSURE_TIME = [float(x) for x in config[10].split(",")]
FILTER_LAMP_INTENSITY = [
    int(x) for x in config[11].split(",")
]  # [1,100,150,254,254,0,150,1]

MIRRORED = False  # If image can be rotated so that North is Up and East is Right, this should be True to ensure that a WCS solution can be found.

SAVE_IMAGES = True  # If True, each image will be saved to SAVE_DATA_PATH
# SAVE_DATA_PATH = r"{MyDocuments}\Calibration\{Timestamp}"
SAVE_DATA_PATH = config[0]  # Script data will be saved to this location.
MASTER_DATA_PATH = config[12]

print(SAVE_DATA_PATH)


### END CONFIGURATION VALUES ########################


def main():
    """
    Starting point for the center_target_on_pixel script
    """

    print("Launching mount control software...")
    mount = Dispatch(MOUNT_DRIVER)

    print("Connecting to mount...")
    mount.Connected = True

    print("Launching MaxIm DL...")
    maxim = Dispatch("MaxIm.Application")
    maxim.LockApp = True

    camera = Dispatch("MaxIm.CCDCamera")
    camera.DisableAutoShutdown = True

    print("Connecting to camera...")
    camera.LinkEnabled = True

    if HOME_MOUNT:
        print("Homing Mount")
        mount.FindHome

    if mount.CanSetTracking:
        # Not all mount drivers support turning tracking on/off.
        # For example, the ASCOM driver for TheSky does not support it.
        # However, if it is supported, make sure tracking is on before slewing
        mount.Tracking = False

    save_data_path = parse_filepath_template(SAVE_DATA_PATH)
    master_data_path = parse_filepath_template(MASTER_DATA_PATH)

    if SAVE_IMAGES and not os.path.isdir(save_data_path):
        print("Creating directory %s" % save_data_path)
        os.makedirs(save_data_path)

    print("Slewing to Azimuth %s, Altitude %s" % (TARGET_AZ, TARGET_ALT))

    mount.SlewToAltAz(TARGET_AZ, TARGET_ALT)
    mount.Tracking = False
    print("Settling...")
    time.sleep(SETTLE_TIME_SECONDS)

    darknum = 0
    print(
        "Taking %d dark images with %d second exposure..."
        % (NUMBER_DARKS, EXPOSURE_LENGTH_SECONDS_DARK)
    )
    while darknum < NUMBER_DARKS:
        camera.BinX = BINNING
        camera.BinY = BINNING
        camera.Expose(
            EXPOSURE_LENGTH_SECONDS_DARK, 0
        )  # The 0 indicates that the shutter is closed
        while not camera.ImageReady:
            time.sleep(0.1)

        darknum = darknum + 1
        print("Dark %d of %d complete" % (darknum, NUMBER_DARKS))

        tempfilename = os.path.join(tempfile.gettempdir(), "Dark.fits")
        camera.SaveImage(tempfilename)

        if SAVE_IMAGES:
            filename = "Dark_%03d.fits" % (darknum)
            # filename = "Dark.fits"
            filepath = os.path.join(save_data_path, filename)
            # print "Saving image to", filepath
            camera.SaveImage(filepath)

    biasnum = 0
    print("Taking %d bias images..." % (NUMBER_BIAS))
    while biasnum < NUMBER_BIAS:
        camera.BinX = BINNING
        camera.BinY = BINNING
        camera.Expose(0, 0)  # Bias image take as a 0 second dark exposure
        while not camera.ImageReady:
            time.sleep(0.1)

        biasnum = biasnum + 1
        print("Bias %d of %d complete" % (biasnum, NUMBER_BIAS))

        tempfilename = os.path.join(tempfile.gettempdir(), "Bias.fits")
        camera.SaveImage(tempfilename)

        if SAVE_IMAGES:
            filename = "Bias_%03d.fits" % (biasnum)
            # filename = "Bias.fits"
            filepath = os.path.join(save_data_path, filename)
            # print "Saving image to", filepath
            camera.SaveImage(filepath)

    filternum = 0
    for fname in FILTER_NAMES:
        # camera.set_active_filter(filternum)
        if FILTER_EXPOSURE_TIME[filternum] == 0:
            print("Skipping filter %s." % (fname))
        else:
            flatnum = 0
            print(
                "Setting flat field lamps to brightness level %d"
                % (FILTER_LAMP_INTENSITY[filternum])
            )
            pulsar_dimmer.dimmer(FILTER_LAMP_INTENSITY[filternum])
            time.sleep(2)
            print(
                "Taking %d flat images with %d second exposure in filter %s..."
                % (NUMBER_FLATS, FILTER_EXPOSURE_TIME[filternum], fname)
            )
            while flatnum < NUMBER_FLATS:
                camera.BinX = BINNING
                camera.BinY = BINNING
                camera.Expose(FILTER_EXPOSURE_TIME[filternum], 1, filternum)
                while not camera.ImageReady:
                    time.sleep(0.1)
                flatnum = flatnum + 1
                print(
                    "Flat %d of %d for filter %s complete"
                    % (flatnum, NUMBER_FLATS, fname)
                )
                camera.SetFITSKey("IMAGETYP", "FLAT    ")
                tempfilename = os.path.join(tempfile.gettempdir(), "Flat.fits")
                camera.SaveImage(tempfilename)

                if SAVE_IMAGES:
                    filename = "Flat_%03d%s.fits" % (flatnum, fname)
                    filepath = os.path.join(save_data_path, filename)
                    # print "Saving image to", filepath
                    camera.SaveImage(filepath)
        filternum = filternum + 1
        print("30 second cooldown...")
        time.sleep(30)
    print("Turning off calibration lamps...")
    pulsar_dimmer.dimmer(0)  # Turn off the flat field lamps when finished exposing


"""
    print "Now median averaging the dark frames..."
    darknum=0
    while darknum < NUMBER_DARKS-1:
        darknum = darknum+1
        filename = "Dark_%03d.fits" % (darknum)
        filepath = os.path.join(save_data_path,filename)
        print "Loading dark image %s" % (filename)

        # if not 'dark_stack' in locals():
        #     hdu_list = fits.open(filepath)
        #     prihdr = hdu_list[0].header
        #     dark_stack = fits.getdata(filepath)
        #     #dark_stack = fits.getdata(filepath)
        # else:
        #     dark_stack = numpy.dstack((dark_stack,fits.getdata(filepath)))

        hdu_list = fits.open(filepath)
        prihdr = hdu_list[0].header
        dark_image = fits.getdata(filepath)
        if not 'dark_stack' in locals():
            dark_stack = numpy.zeros(shape=dark_image.shape + (NUMBER_DARKS,))
        dark_stack[:,:,darknum]


    if not darknum == 0:
        dark_master = numpy.median(dark_stack, axis=2)
        prihdr['IMAGETYP'] = 'DARK'
        hdu = fits.PrimaryHDU(dark_master,prihdr)
        darkhdulist = fits.HDUList([hdu])
        filename = "Master_Dark.fits"
        filepath = os.path.join(save_data_path,filename)
        print "Saving master dark image"
        darkhdulist.writeto(filepath)
        filepath = os.path.join(master_data_path,filename)
        darkhdulist.writeto(filepath,clobber=True)
        darkhdulist.close()
        hdu_list.close()
        dark_stack=None
        dark_master=None

    print "Now median averaging the bias frames..."
    biasnum=0
    while biasnum < NUMBER_BIAS-1:
        biasnum = biasnum+1
        filename = "Bias_%03d.fits" % (biasnum)
        filepath = os.path.join(save_data_path,filename)
        print "Loading bias image %s" % (filename)
        hdu_list = fits.open(filepath)
        prihdr = hdu_list[0].header
        bias_image = fits.getdata(filepath)
        if not 'bias_stack' in locals():
            bias_stack = numpy.zeros(shape=bias_image.shape + (NUMBER_BIAS,))
        bias_stack[:,:,biasnum]
        # if not 'bias_stack' in locals():
        #     hdu_list = fits.open(filepath)
        #     prihdr = hdu_list[0].header
        #     hdu_list.close()
        #     bias_stack = fits.getdata(filepath)
        #     #dark_stack = fits.getdata(filepath)
        # else:
        #     bias_stack = numpy.dstack((bias_stack,fits.getdata(filepath)))
    if not biasnum == 0:
        bias_master = numpy.median(bias_stack, axis=2)
        prihdr['IMAGETYP'] = 'BIAS    '
        hdu = fits.PrimaryHDU(bias_master,prihdr)
        biashdulist = fits.HDUList([hdu])
        filename = "Master_Bias.fits"
        filepath = os.path.join(save_data_path,filename)
        print "Saving master Bias image"
        biashdulist.writeto(filepath)
        filepath = os.path.join(master_data_path,filename)
        biashdulist.writeto(filepath,clobber=True)
        biashdulist.close()

        bias_stack=None
        bias_master=None

    filternum=0
    for fname in FILTER_NAMES:
        if FILTER_EXPOSURE_TIME[filternum] == 0:
            print "Skiping filter %s." % (fname)
        else:
            flatnum=0
            print "Now median averaging the flat frames for the %s filter..." % (fname)
            while flatnum < NUMBER_FLATS:
                flatnum = flatnum+1
                filename = "Flat_%03d%s.fits" % (flatnum,fname)
                filepath = os.path.join(save_data_path,filename)
                print "Loading flat image %s" % (filename)

                hdu_list = fits.open(filepath)
                prihdr = hdu_list[0].header
                flat_image = fits.getdata(filepath)
                if not 'flat_stack' in locals():
                    flat_stack = numpy.zeros(shape=flat_image.shape + (NUMBER_FLATS,))
                #if flat_stack is None:
                #    flat_stack = numpy.zeros(shape=flat_image.shape + (NUMBER_FLATS,))

                flat_stack[:,:,flatnum-1]

                # if flatnum==1:
                #     hdu_list = fits.open(filepath)
                #     prihdr = hdu_list[0].header
                #     flat_stack = fits.getdata(filepath)
                # else:
                #     flat_stack = numpy.dstack((flat_stack,fits.getdata(filepath)))
            if not flatnum == 0:
                flat_master = numpy.median(flat_stack, axis=2)
                prihdr['IMAGETYP'] = 'FLAT    '
                hdu = fits.PrimaryHDU(flat_master,prihdr)

                flathdulist = fits.HDUList([hdu])
                filename = "Master_Flat_%s.fits" % (fname)
                filepath = os.path.join(save_data_path,filename)
                print "Saving master flat image for filter %s" % (fname)
                flathdulist.writeto(filepath)
                filepath = os.path.join(master_data_path,filename)
                flathdulist.writeto(filepath,clobber=True)
                flathdulist.close()
                hdu_list.close()

        filternum=filternum+1
"""


def parse_filepath_template(template):
    my_documents_path = os.path.expanduser(r"~\My Documents")
    timestamp = datetime.now().strftime("%Y-%m-%d %H_%M_%S")

    template = template.replace("{MyDocuments}", my_documents_path)
    template = template.replace("{Timestamp}", timestamp)

    return template


if __name__ == "__main__":
    main()



import platform
import time

import pytest

from pyscope.observatory import Maxim


@pytest.mark.skipif(platform.system() != "Windows", reason="Maxim only runs on Windows")
def test_maxim():
    m = Maxim()
    assert m.Connected
    assert m.name == "MaxIm DL"
    assert m.app is not None
    assert m.autofocus is not None
    m.Connected = False


@pytest.mark.skipif(platform.system() != "Windows", reason="Maxim only runs on Windows")
def test_maxim_autofocus():
    m = Maxim()
    success = m.autofocus.Run(exposure=1)
    assert success is not None

    with pytest.raises(NotImplementedError):
        m.autofocus.Abort()

    m.Connected = False


@pytest.mark.skipif(platform.system() != "Windows", reason="Maxim only runs on Windows")
def test_maxim_camera():
    m = Maxim()
    c = m.camera
    c.Connected = True
    assert c.Connected
    assert c.Name is not None

    c.StartExposure(1, True)
    time.sleep(0.5)
    if c.CanStopExposure:
        c.StopExposure()

    c.StartExposure(1, False)
    time.sleep(0.5)
    if c.CanAbortExposure:
        c.AbortExposure()

    with pytest.raises(NotImplementedError):
        c.PulseGuide(0, 0)

    with pytest.raises(NotImplementedError):
        c.BayerOffsetX = 0

    with pytest.raises(NotImplementedError):
        c.BayerOffsetY = 0

    c.BinX = 2
    assert c.BinX == 2
    if not c.CanAsymmetricBin:
        assert c.BinY == 2
    else:
        assert c.BinY == 1

    with pytest.raises(NotImplementedError):
        c.CanFastReadout

    assert c.CameraState is not None

    assert c.CameraXSize is not None
    assert c.CameraYSize is not None

    assert c.CanGetCoolerPower

    assert not c.CanPulseGuide

    if c.CanSetCCDTemperature:
        c.SetCCDTemperature = -20
        assert c.SetCCDTemperature == -20

    assert c.CCDTemperature is not None

    c.CoolerOn = True
    assert c.CoolerOn
    assert c.CoolerPower is not None
    c.CoolerOn = False
    assert not c.CoolerOn

    with pytest.raises(NotImplementedError):
        c.ElectronsPerADU
    with pytest.raises(NotImplementedError):
        c.ExposureMax
    with pytest.raises(NotImplementedError):
        c.ExposureMin
    with pytest.raises(NotImplementedError):
        c.ExposureResolution

    with pytest.raises(DeprecationWarning):
        c.FastReadout
    with pytest.raises(DeprecationWarning):
        c.FastReadout = True

    with pytest.raises(NotImplementedError):
        c.FullWellCapacity

    c.Gain = c.Gains[0]
    assert c.Gain == c.Gains[0]

    with pytest.raises(NotImplementedError):
        c.GainMax
    with pytest.raises(NotImplementedError):
        c.GainMin

    assert c.HasShutter is not None
    assert c.HeatSinkTemperature is not None

    c.StartExposure(0.1, True)
    while not c.ImageReady:
        time.sleep(0.1)
    assert c.ImageArray is not None

    with pytest.raises(NotImplementedError):
        c.IsPulseGuiding

    assert c.LastExposureDuration is not None
    assert c.LastExposureStartTime is not None

    with pytest.raises(NotImplementedError):
        c.MaxADU

    assert c.MaxBinX is not None
    assert c.MaxBinY is not None

    c.NumX = 100
    assert c.NumX == 100
    c.NumY = 100
    assert c.NumY == 100

    with pytest.raises(NotImplementedError):
        c.Offset
    with pytest.raises(NotImplementedError):
        c.Offset = 1
    with pytest.raises(NotImplementedError):
        c.OffsetMax
    with pytest.raises(NotImplementedError):
        c.OffsetMin
    with pytest.raises(NotImplementedError):
        c.Offsets
    with pytest.raises(NotImplementedError):
        c.PercentCompleted

    assert c.PixelSizeX is not None
    assert c.PixelSizeY is not None

    c.ReadoutMode = 0
    assert c.ReadoutMode == 0
    assert c.ReadoutModes is not None

    with pytest.raises(NotImplementedError):
        c.SensorName
    with pytest.raises(NotImplementedError):
        c.SensorType

    c.StartX = 1
    assert c.StartX == 1
    c.StartY = 1
    assert c.StartY == 1

    with pytest.raises(NotImplementedError):
        c.SubExposureDuration
    with pytest.raises(NotImplementedError):
        c.SubExposureDuration = 1

    m.Connected = False


@pytest.mark.skipif(platform.system() != "Windows", reason="Maxim only runs on Windows")
def test_maxim_filter_wheel():
    m = Maxim()
    f = m.filter_wheel
    f.Connected = True
    assert f.Connected

    assert f.Name is not None

    with pytest.raises(NotImplementedError):
        f.FocusOffsets

    assert f.Names is not None
    assert f.Position is not None
    f.Position = 1
    assert f.Position == 1

    m.Connected = False


@pytest.mark.skipif(platform.system() != "Windows", reason="Maxim only runs on Windows")
def test_maxim_pinpoint():
    m = Maxim()
    p = m.pinpoint

    m.Camera.StartExposure(1, True)
    while not m.Camera.ImageReady:
        time.sleep(0.1)

    assert p.Solve() is not None

    m.Connected = False



from abc import ABC, abstractmethod

from ._docstring_inheritee import _DocstringInheritee


class ObservingConditions(ABC, metaclass=_DocstringInheritee):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def Refresh(self):
        pass

    @abstractmethod
    def SensorDescription(self, PropertyName):
        pass

    @abstractmethod
    def TimeSinceLastUpdate(self, PropertyName):
        pass

    @property
    @abstractmethod
    def AveragePeriod(self):
        pass

    @AveragePeriod.setter
    @abstractmethod
    def AveragePeriod(self, value):
        pass

    @property
    @abstractmethod
    def CloudCover(self):
        pass

    @property
    @abstractmethod
    def DewPoint(self):
        pass

    @property
    @abstractmethod
    def Humidity(self):
        pass

    @property
    @abstractmethod
    def Pressure(self):
        pass

    @property
    @abstractmethod
    def RainRate(self):
        pass

    @property
    @abstractmethod
    def SkyBrightness(self):
        pass

    @property
    @abstractmethod
    def SkyQuality(self):
        pass

    @property
    @abstractmethod
    def SkyTemperature(self):
        pass

    @property
    @abstractmethod
    def StarFWHM(self):
        pass

    @property
    @abstractmethod
    def Temperature(self):
        pass

    @property
    @abstractmethod
    def WindDirection(self):
        pass

    @property
    @abstractmethod
    def WindGust(self):
        pass

    @property
    @abstractmethod
    def WindSpeed(self):
        pass



import datetime
import smtplib

gmail_user = "iota.alert@gmail.com"
gmail_password = "iowaTele1"

sent_from = gmail_user
to = ["chris@iapcrepair.com", "robert-mutel@uiowa.edu", "caroline-roberts@uiowa.edu"]
subject = "Gemini ECC PC Rebooted"
body = "Gemini ECC PC Rebooted at " + str(datetime.datetime.now())

email_text = """\
From: %s
To: %s
Subject: %s

%s
""" % (
    sent_from,
    ", ".join(to),
    subject,
    body,
)

try:
    server = smtplib.SMTP_SSL("smtp.gmail.com", 465)
    server.ehlo()
    server.login(gmail_user, gmail_password)
    server.sendmail(sent_from, to, email_text)
    server.close()

    print("Email sent!")
except:
    print("Something went wrong...")



import logging

from .ascom_device import ASCOMDevice
from .rotator import Rotator

logger = logging.getLogger(__name__)


class ASCOMRotator(ASCOMDevice, Rotator):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    def Halt(self):
        logger.debug("ASCOMRotator.Halt() called")
        self._device.Halt()

    def Move(self, Position):
        logger.debug(f"ASCOMRotator.Move({Position}) called")
        self._device.Move(Position)

    def MoveAbsolute(self, Position):
        logger.debug(f"ASCOMRotator.MoveAbsolute({Position}) called")
        self._device.MoveAbsolute(Position)

    def MoveMechanical(self, Position):
        logger.debug(f"ASCOMRotator.MoveMechanical({Position}) called")
        self._device.MoveMechanical(Position)

    def Sync(self, Position):
        logger.debug(f"ASCOMRotator.Sync({Position}) called")
        self._device.Sync(Position)

    @property
    def CanReverse(self):
        logger.debug("ASCOMRotator.CanReverse property called")
        return self._device.CanReverse

    @property
    def IsMoving(self):
        logger.debug("ASCOMRotator.IsMoving property called")
        return self._device.IsMoving

    @property
    def MechanicalPosition(self):
        logger.debug("ASCOMRotator.MechanicalPosition property called")
        return self._device.MechanicalPosition

    @property
    def Position(self):
        logger.debug("ASCOMRotator.Position property called")
        return self._device.Position

    @property
    def Reverse(self):
        logger.debug("ASCOMRotator.Reverse property called")
        return self._device.Reverse

    @Reverse.setter
    def Reverse(self, value):
        logger.debug(f"ASCOMRotator.Reverse property set to {value}")
        self._device.Reverse = value

    @property
    def StepSize(self):
        logger.debug("ASCOMRotator.StepSize property called")
        return self._device.StepSize

    @property
    def TargetPosition(self):
        logger.debug("ASCOMRotator.TargetPosition property called")
        return self._device.TargetPosition



import logging

from .ascom_device import ASCOMDevice
from .camera import Camera

logger = logging.getLogger(__name__)


class ASCOMCamera(ASCOMDevice, Camera):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    def AbortExposure(self):
        logger.debug(f"ASCOMCamera.AbortExposure() called")
        self._device.AbortExposure()

    def PulseGuide(self, Direction, Duration):
        logger.debug(f"ASCOMCamera.PulseGuide({Direction}, {Duration}) called")
        self._device.PulseGuide(Direction, Duration)

    def StartExposure(self, Duration, Light):
        logger.debug(f"ASCOMCamera.StartExposure({Duration}, {Light}) called")
        self._device.StartExposure(Duration, Light)

    def StopExposure(self):
        logger.debug(f"ASCOMCamera.StopExposure() called")
        self._device.StopExposure()

    @property
    def BayerOffsetX(self):  # pragma: no cover
        """
        .. warning::
            This property is not implemented in the ASCOM Alpaca protocol.
        """
        logger.debug(f"ASCOMCamera.BayerOffsetX property called")
        return self._device.BayerOffsetX

    @property
    def BayerOffsetY(self):  # pragma: no cover
        """
        .. warning::
            This property is not implemented in the ASCOM Alpaca protocol.
        """
        logger.debug(f"ASCOMCamera.BayerOffsetY property called")
        return self._device.BayerOffsetY

    @property
    def BinX(self):
        logger.debug(f"ASCOMCamera.BinX property called")
        return self._device.BinX

    @BinX.setter
    def BinX(self, value):
        logger.debug(f"ASCOMCamera.BinX property set to {value}")
        self._device.BinX = value

    @property
    def BinY(self):
        logger.debug(f"ASCOMCamera.BinY property called")
        return self._device.BinY

    @BinY.setter
    def BinY(self, value):
        logger.debug(f"ASCOMCamera.BinY property set to {value}")
        self._device.BinY = value

    @property
    def CameraState(self):
        logger.debug(f"ASCOMCamera.CameraState property called")
        return self._device.CameraState

    @property
    def CameraXSize(self):
        logger.debug(f"ASCOMCamera.CameraXSize property called")
        return self._device.CameraXSize

    @property
    def CameraYSize(self):
        logger.debug(f"ASCOMCamera.CameraYSize property called")
        return self._device.CameraYSize

    @property
    def CanAbortExposure(self):
        logger.debug(f"ASCOMCamera.CanAbortExposure property called")
        return self._device.CanAbortExposure

    @property
    def CanAsymmetricBin(self):
        logger.debug(f"ASCOMCamera.CanAsymmetricBin property called")
        return self._device.CanAsymmetricBin

    @property
    def CanFastReadout(self):
        logger.debug(f"ASCOMCamera.CanFastReadout property called")
        return self._device.CanFastReadout

    @property
    def CanGetCoolerPower(self):
        logger.debug(f"ASCOMCamera.CanGetCoolerPower property called")
        return self._device.CanGetCoolerPower

    @property
    def CanPulseGuide(self):
        logger.debug(f"ASCOMCamera.CanPulseGuide property called")
        return self._device.CanPulseGuide

    @property
    def CanSetCCDTemperature(self):
        logger.debug(f"ASCOMCamera.CanSetCCDTemperature property called")
        return self._device.CanSetCCDTemperature

    @property
    def CanStopExposure(self):
        logger.debug(f"ASCOMCamera.CanStopExposure property called")
        return self._device.CanStopExposure

    @property
    def CCDTemperature(self):
        logger.debug(f"ASCOMCamera.CCDTemperature property called")
        return self._device.CCDTemperature

    @property
    def CoolerOn(self):
        logger.debug(f"ASCOMCamera.CoolerOn property called")
        return self._device.CoolerOn

    @CoolerOn.setter
    def CoolerOn(self, value):
        logger.debug(f"ASCOMCamera.CoolerOn property set to {value}")
        self._device.CoolerOn = value

    @property
    def CoolerPower(self):
        logger.debug(f"ASCOMCamera.CoolerPower property called")
        return self._device.CoolerPower

    @property
    def ElectronsPerADU(self):
        logger.debug(f"ASCOMCamera.ElectronsPerADU() property called")
        return self._device.ElectronsPerADU

    @property
    def ExposureMax(self):
        logger.debug(f"ASCOMCamera.ExposureMax property called")
        return self._device.ExposureMax

    @property
    def ExposureMin(self):
        logger.debug(f"ASCOMCamera.ExposureMin property called")
        return self._device.ExposureMin

    @property
    def ExposureResolution(self):
        logger.debug(f"ASCOMCamera.ExposureResolution property called")
        return self._device.ExposureResolution

    @property
    def FastReadout(self):
        logger.debug(f"ASCOMCamera.FastReadout property called")
        return self._device.FastReadout

    @FastReadout.setter
    def FastReadout(self, value):
        logger.debug(f"ASCOMCamera.FastReadout property set to {value}")
        self._device.FastReadout = value

    @property
    def FullWellCapacity(self):
        logger.debug(f"ASCOMCamera.FullWellCapacity property called")
        return self._device.FullWellCapacity

    @property
    def Gain(self):
        logger.debug(f"ASCOMCamera.Gain property called")
        return self._device.Gain

    @Gain.setter
    def Gain(self, value):
        logger.debug(f"ASCOMCamera.Gain property set to {value}")
        self._device.Gain = value

    @property
    def GainMax(self):
        logger.debug(f"ASCOMCamera.GainMax property called")
        return self._device.GainMax

    @property
    def GainMin(self):
        logger.debug(f"ASCOMCamera.GainMin property called")
        return self._device.GainMin

    @property
    def Gains(self):
        logger.debug(f"ASCOMCamera.Gains property called")
        return self._device.Gains

    @property
    def HasShutter(self):
        logger.debug(f"ASCOMCamera.HasShutter property called")
        return self._device.HasShutter

    @property
    def HeatSinkTemperature(self):
        logger.debug(f"ASCOMCamera.HeatSinkTemperature property called")
        return self._device.HeatSinkTemperature

    @property
    def ImageArray(self):
        logger.debug(f"ASCOMCamera.ImageArray property called")
        return self._device.ImageArray

    @property
    def ImageReady(self):
        logger.debug(f"ASCOMCamera.ImageReady property called")
        return self._device.ImageReady

    @property
    def IsPulseGuiding(self):
        logger.debug(f"ASCOMCamera.IsPulseGuiding property called")
        return self._device.IsPulseGuiding

    @property
    def LastExposureDuration(self):
        logger.debug(f"ASCOMCamera.LastExposureDuration property called")
        return self._device.LastExposureDuration

    @property
    def LastExposureStartTime(self):
        logger.debug(f"ASCOMCamera.LastExposureStartTime property called")
        return self._device.LastExposureStartTime

    @property
    def MaxADU(self):
        logger.debug(f"ASCOMCamera.MaxADU property called")
        return self._device.MaxADU

    @property
    def MaxBinX(self):
        logger.debug(f"ASCOMCamera.MaxBinX property called")
        return self._device.MaxBinX

    @property
    def MaxBinY(self):
        logger.debug(f"ASCOMCamera.MaxBinY property called")
        return self._device.MaxBinY

    @property
    def NumX(self):
        logger.debug(f"ASCOMCamera.NumX property called")
        return self._device.NumX

    @NumX.setter
    def NumX(self, value):
        logger.debug(f"ASCOMCamera.NumX property set to {value}")
        self._device.NumX = value

    @property
    def NumY(self):
        logger.debug(f"ASCOMCamera.NumY property called")
        return self._device.NumY

    @NumY.setter
    def NumY(self, value):
        logger.debug(f"ASCOMCamera.NumY property set to {value}")
        self._device.NumY = value

    @property
    def Offset(self):
        logger.debug(f"ASCOMCamera.Offset property called")
        return self._device.Offset

    @Offset.setter
    def Offset(self, value):
        logger.debug(f"ASCOMCamera.Offset property set to {value}")
        self._device.Offset = value

    @property
    def OffsetMax(self):
        logger.debug(f"ASCOMCamera.OffsetMax property called")
        return self._device.OffsetMax

    @property
    def OffsetMin(self):
        logger.debug(f"ASCOMCamera.OffsetMin property called")
        return self._device.OffsetMin

    @property
    def Offsets(self):
        logger.debug(f"ASCOMCamera.Offsets property called")
        return self._device.Offsets

    @property
    def PercentCompleted(self):
        logger.debug(f"ASCOMCamera.PercentCompleted property called")
        return self._device.PercentCompleted

    @property
    def PixelSizeX(self):
        logger.debug(f"ASCOMCamera.PixelSizeX property called")
        return self._device.PixelSizeX

    @property
    def PixelSizeY(self):
        logger.debug(f"ASCOMCamera.PixelSizeY property called")
        return self._device.PixelSizeY

    @property
    def ReadoutMode(self):
        logger.debug(f"ASCOMCamera.ReadoutMode property called")
        return self._device.ReadoutMode

    @ReadoutMode.setter
    def ReadoutMode(self, value):
        logger.debug(f"ASCOMCamera.ReadoutMode property set to {value}")
        self._device.ReadoutMode = value

    @property
    def ReadoutModes(self):
        logger.debug(f"ASCOMCamera.ReadoutModes property called")
        return self._device.ReadoutModes

    @property
    def SensorName(self):
        logger.debug(f"ASCOMCamera.SensorName property called")
        return self._device.SensorName

    @property
    def SensorType(self):
        logger.debug(f"ASCOMCamera.SensorType property called")
        return self._device.SensorType

    @property
    def SetCCDTemperature(self):
        logger.debug(f"ASCOMCamera.SetCCDTemperature property called")
        return self._device.SetCCDTemperature

    @SetCCDTemperature.setter
    def SetCCDTemperature(self, value):
        logger.debug(f"ASCOMCamera.SetCCDTemperature property set to {value}")
        self._device.SetCCDTemperature = value

    @property
    def StartX(self):
        logger.debug(f"ASCOMCamera.StartX property called")
        return self._device.StartX

    @StartX.setter
    def StartX(self, value):
        logger.debug(f"ASCOMCamera.StartX property set to {value}")
        self._device.StartX = value

    @property
    def StartY(self):
        logger.debug(f"ASCOMCamera.StartY property called")
        return self._device.StartY

    @StartY.setter
    def StartY(self, value):
        logger.debug(f"ASCOMCamera.StartY property set to {value}")
        self._device.StartY = value

    @property
    def SubExposureDuration(self):
        logger.debug(f"ASCOMCamera.SubExposureDuration property called")
        return self._device.SubExposureDuration

    @SubExposureDuration.setter
    def SubExposureDuration(self, value):
        logger.debug(f"ASCOMCamera.SubExposureDuration property set to {value}")
        self._device.SubExposureDuration = value



import logging

from .ascom_device import ASCOMDevice
from .observing_conditions import ObservingConditions

logger = logging.getLogger(__name__)


class ASCOMObservingConditions(ASCOMDevice, ObservingConditions):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    def Refresh(self):
        logger.debug("ASCOMObservingConditions.Refresh() called")
        self._device.Refresh()

    def SensorDescription(self, PropertyName):
        logger.debug(
            f"ASCOMObservingConditions.SensorDescription({PropertyName}) called"
        )
        return self._device.SensorDescription(PropertyName)

    def TimeSinceLastUpdate(self, PropertyName):
        logger.debug(
            f"ASCOMObservingConditions.TimeSinceLastUpdate({PropertyName}) called"
        )
        return self._device.TimeSinceLastUpdate(PropertyName)

    @property
    def AveragePeriod(self):
        logger.debug("ASCOMObservingConditions.AveragePeriod property called")
        return self._device.AveragePeriod

    @AveragePeriod.setter
    def AveragePeriod(self, value):
        logger.debug(f"ASCOMObservingConditions.AveragePeriod property set to {value}")
        self._device.AveragePeriod = value

    @property
    def CloudCover(self):
        logger.debug("ASCOMObservingConditions.CloudCover property called")
        return self._device.CloudCover

    @property
    def DewPoint(self):
        logger.debug("ASCOMObservingConditions.DewPoint property called")
        return self._device.DewPoint

    @property
    def Humidity(self):
        logger.debug("ASCOMObservingConditions.Humidity property called")
        return self._device.Humidity

    @property
    def Pressure(self):
        logger.debug("ASCOMObservingConditions.Pressure property called")
        return self._device.Pressure

    @property
    def RainRate(self):
        logger.debug("ASCOMObservingConditions.RainRate property called")
        return self._device.RainRate

    @property
    def SkyBrightness(self):
        logger.debug("ASCOMObservingConditions.SkyBrightness property called")
        return self._device.SkyBrightness

    @property
    def SkyQuality(self):
        logger.debug("ASCOMObservingConditions.SkyQuality property called")
        return self._device.SkyQuality

    @property
    def SkyTemperature(self):
        logger.debug("ASCOMObservingConditions.SkyTemperature property called")
        return self._device.SkyTemperature

    @property
    def StarFWHM(self):
        logger.debug("ASCOMObservingConditions.StarFWHM property called")
        return self._device.StarFWHM

    @property
    def Temperature(self):
        logger.debug("ASCOMObservingConditions.Temperature property called")
        return self._device.Temperature

    @property
    def WindDirection(self):
        logger.debug("ASCOMObservingConditions.WindDirection property called")
        return self._device.WindDirection

    @property
    def WindGust(self):
        logger.debug("ASCOMObservingConditions.WindGust property called")
        return self._device.WindGust

    @property
    def WindSpeed(self):
        logger.debug("ASCOMObservingConditions.WindSpeed property called")
        return self._device.WindSpeed



import logging
import platform
import time

from .autofocus import Autofocus

logger = logging.getLogger(__name__)


class PWIAutofocus(Autofocus):
    def __init__(self):
        logger.debug("PWIAutofocus.__init__ called")
        if platform.system() != "Windows":
            raise Exception("This class is only available on Windows.")
        else:
            from win32com.client import Dispatch

            self._com_object = Dispatch("PlaneWave.AutoFocus")

            self._com_object.StartPwiIfNeeded
            self._forward_autofocus_messages()
            self._com_object.ConnectFocuser
            self._forward_autofocus_messages()

            # Wait up to 3 seconds for a confirmed connection
            t = time.time() + 3
            while not self._com_object.IsFocuserConnected:
                if time.time() > t:
                    raise Exception("Unable to connect to focuser.")
                time.sleep(0.1)
                self._forward_autofocus_messages()

            self._com_object.PreventFilterChange = True

    def Run(self, exposure=10, timeout=120):
        logger.debug(
            f"PWIAutofocus.Run called with args: exposure={exposure}, timeout={timeout}"
        )
        self._com_object.ExposureLengthSeconds = exposure

        if not self._com_object.IsFocuserConnected:
            raise Exception(
                "Unable to run PlaneWave AutoFocus: focuser is not connected"
            )

        self._com_object.StartAutoFocus

        t = time.time() + timeout
        while self._com_object.IsAutoFocusRunning:
            self._forward_autofocus_messages()
            time.sleep(0.2)

            if time.time() > t:
                raise Exception(
                    "Autofocus took longer than %g seconds to complete" % timeout
                )

        if self._com_object.Success:
            return self._com_object.BestPosition
        else:
            return None

    def Abort(self):
        logger.debug("PWIAutofocus.Abort called")
        self._com_object.StopAutofocus

    def _forward_autofocus_messages(self):
        while True:
            log_line = _autofocus.NextLogMessage
            if log_line is None:
                return
            logger.info(log_line)



import logging

import click
import markdown
import numpy as np
from astropy import coordinates as coord
from astropy import table
from astropy import time as astrotime
from astropy.io import fits

from ..reduction import fitslist

logger = logging.getLogger(__name__)


@click.command(
    epilog="""Check out the documentation at
               https://pyscope.readthedocs.io/en/latest/
               for more information."""
)
@click.option(
    "-s",
    "--schedule",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="""Path to the schedule file. If none provided, the default is in
    the ./schedules/ directory, find the latest schedule.""",
)
@click.option(
    "-i",
    "--images-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    default="./images/",
    show_default=True,
    help="Path to the directory containing the images.",
)
@click.option(
    "-p",
    "--preamble",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="""Path to the preamble file. Must be a .md file, which will be displayed
    prior to the summary information.""",
)
@click.option(
    "-c",
    "--code",
    type=str,
    help="""The observing code to generate the report for. If none provided,
    a report including all observing codes will be generated.""",
)
@click.option(
    "-o",
    "--output",
    default="",
    type=click.Path(exists=False, dir_okay=False, resolve_path=True),
    help="""Path to the output file. If none provided, the default is report_%Y-%m-%d
    where the date is from the first observation in the schedules directory.""",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    type=click.IntRange(0, 1),  # Range can be changed
    default=0,
    help="Increase verbosity",
)
@click.version_option()
def summary_report_cli(
    schedule, images_dir="./images/", preamble=None, code=None, verbose=-1
):
    """
    Generates an HTML summary report on the requested schedule.\b

    Using information written to the schedule file by TelrunOperator and
    the corresponding image files from that schedule, a summary report
    detailing the observing conditions and the specific outcomes of each
    image is generated. The report is written to a .html file for easy
    viewing in a web browser and compatability with various summary
    delivery methods (e.g. email, webpage, etc.).

    .. hint::
        A unique report could be generated for each observing code by
        using the ``--code`` option. For example, each observer could
        receive a custom report detailing only their observations.

    Parameters
    ----------
    schedule : `str`, optional
        Path to the schedule file. If none provided, the default behavior
        is to check the ./schedules/ directory for the latest schedule.

    images_dir : `str`, default='./images/'
        Path to the directory containing the images.

    preamble : `str`, optional
        Path to the preamble file. Must be a .md file, which will be
        displayed prior to the summary information.

    code : `str`, optional
        The observing code to generate the report for. If none provided,
        a report including all observing codes will be generated.

    output : `str`, optional
        Path to the output file. If none provided from the CLI, the default is in the
        ./reports/ directory, report_%Y-%m-%d where the date is from the
        first observation's UTC time in the schedules directory. If none
        provided from the API, the output HTML is returned as a string.

    verbose : `int`, {-1, 0, 1}, default=-1
        Verbosity level. 0: no debug messages, 1: debug messages. Default set to -1
        for API use and 0 for CLI use.

    Returns
    -------
    `str`, optional
        If no output file is provided, the HTML report is returned as a string.

    Raises
    ------
    `click.BadOptionUsage`

    See Also
    --------


    Notes
    -----


    Examples
    --------

    """

    if verbose > -1:
        logger.setLevel(int(10 * (2 - verbose)))  # Change range via 2
        logger.addHandler(logging.StreamHandler())  # Redirect to stdout
    logger.debug(f"Verbosity level set to {verbose}")
    logger.debug(f"""summary_report_cli()""")

    if code is not None:
        if len(code) != 3:
            raise click.BadOptionUsage("Observing codes must be 3 characters long.")

    # Get schedule
    if schedule is None:
        last_mtime = 0
        for fname in glob.glob("./schedules/*.ecsv"):
            if os.path.getmtime(schedule) > last_mtime:
                last_mtime = os.path.getmtime(schedule)
                schedule = fname
    schedule = table.Table.read(schedule, format="ascii.ecsv")

    # Check images all exist
    fnames = []
    headers = []
    for row in schedule:
        for fname in row["configuration"]["filename"].split(","):
            if not os.path.exists(os.path.join(images_dir, fname)):
                raise click.BadOptionUsage(f"Image {fname} does not exist.")
            else:
                fnames.append(os.path.join(images_dir, fname))
                headers.append(fits.getheader(fnames[-1]))

    tbl = table.Table(
        [
            [hdr["OBSCODE"] for hdr in headers],
            [hdr["OBSERVER"] for hdr in headers],
            fnames,
            [hdr["TARGET"] for hdr in headers],
            [
                astrotime.Time(hdr["DATE-OBS"], format="fits", scale="utc")
                for hdr in headers
            ],
            [
                astrotime.Time(row["start time (UTC)"], format="iso", scale="utc")
                for row in schedule
            ],
            [hdr.get("FILTER", "None") for hdr in headers],
            [hdr["READOUT"] for hdr in headers],
            [str(hdr["XBINNING"]) + "x" + str(hdr["YBINNING"]) for hdr in headers],
            [hdr["EXPTIME"] for hdr in headers],
            [
                coord.SkyCoord(
                    hdr["SCHEDRA"], hdr["SCHEDDEC"], unit=("hourangle", "deg")
                )
                for hdr in headers
            ],
            [hdr["AIRMASS"] for hdr in headers],
            [hdr.get("ZMAG", np.nan) for hdr in headers],
            [hdr.get("ZMAGERR", np.nan) for hdr in headers],
            [hdr.get("FWHMH", np.nan) for hdr in headers],
            [hdr.get("FWHMHS", np.nan) for hdr in headers],
            [hdr.get("FWHMV", np.nan) for hdr in headers],
            [hdr.get("FWHMVS", np.nan) for hdr in headers],
            [hdr["XPIXSCAL"] for hdr in headers],
            [hdr["YPIXSCAL"] for hdr in headers],
            [hdr["MOONANGL"] for hdr in headers],
            [hdr["MOONPHAS"] for hdr in headers],
            [
                coord.SkyCoord(
                    hdr["OBJCTRA"], hdr["OBJCTDEC"], unit=("hourangle", "deg")
                )
                if None not in (hdr.get("OBJCTRA", None), hdr.get("OBJCTDEC", None))
                else None
                for hdr in headers
            ],
        ],
        names=[
            "code",
            "observer",
            "filename",
            "target",
            "obs time",
            "start time",
            "filter",
            "readout",
            "binning",
            "exptime",
            "sched coords",
            "airmass",
            "Zmag",
            "Zmag_err",
            "FWHM_H",
            "FWHM_H_err",
            "FWHM_V",
            "FWHM_V_err",
            "xpixscale",
            "ypixscale",
            "moon angle",
            "moon phase",
            "obj coords",
        ],
    )

    tbl.sort("start time")

    # Get summary information
    n_codes = len(np.unique(tbl["code"]))
    n_images = len(tbl)
    exp_tot = np.sum(tbl["exptime"])
    moon_phs_range = (np.min(tbl["moon phase"]), np.max(tbl["moon phase"]))

    # Seeing
    seeing_arr = np.mean(
        [tbl["xpixscale"] * tbl["FWHM_H"], tbl["ypixscale"] * tbl["FWHM_V"]], axis=0
    )
    seeing_median = np.median(seeing_arr)
    seeing_std = np.std(seeing_arr)

    # Zmag
    zmags = {}
    zmag_errs = {}
    for filt in ["u", "b", "g", "r", "i", "z"]:
        reads = {}
        read_errs = {}
        for readout in np.unique(tbl["readout"]):
            idx = np.where(
                (tbl["filter"].lower() == filt) & (tbl["readout"] == readout)
            )[0]
            if len(idx) > 0:
                reads[readout] = np.median(tbl["Zmag"][idx])
                read_errs[readout] = np.std(tbl["Zmag"][idx])
        if len(reads) > 0:
            zmags[filt] = reads
            zmag_errs[filt] = read_errs

    # Pointing errors in arcsec
    dra = np.array(
        [
            (tbl["sched coords"][i].ra.deg - tbl["obj coords"][i].ra.deg) / u.arcsec
            if tbl["obj coords"][i] is not None
            else np.nan
            for i in range(len(tbl))
        ]
    )
    ddec = np.array(
        [
            (tbl["sched coords"][i].dec.deg - tbl["obj coords"][i].dec.deg) / u.arcsec
            if tbl["obj coords"][i] is not None
            else np.nan
            for i in range(len(tbl))
        ]
    )
    dra_median = np.median(dra)
    dra_std = np.std(dra)
    ddec_median = np.median(ddec)
    ddec_std = np.std(ddec)

    # Median late time in minutes
    late_median = np.median((tbl["obs time"] - tbl["start time"]).sec) / 60
    late_std = np.std((tbl["obs time"] - tbl["start time"]).sec) / 60

    # Summary table
    summary_table = table.Table(
        [
            np.unique(tbl["code"]),
            [tbl["observer"][tbl["code"] == code] for code in all_codes],
            [len(tbl[tbl["code"] == code]) for code in all_codes],
            [
                np.sum(tbl["exptime"][tbl["code"] == code]) / 60 for code in all_codes
            ],  # Minutes
            [np.unique(tbl["target"][tbl["code"] == code]) for code in all_codes],
        ],
        names=["code", "observer", "n images", "exp time [min]", "targets"],
    )

    if output == "":
        output = "report_" + tbl["start time"][0].iso.split("T")[0] + ".html"

    # Generate code specific table
    if code is not None:
        tbl = tbl[tbl["code"] == code]

    if output is not None:
        with open(output, "w") as f:
            f.write(
                f"""<!DOCTYPE html>
            <html>
            <head>
            <style>
            table {{
              font-family: arial, sans-serif;
              border-collapse: collapse;
              width: 100%;
            }}

            td, th {{
              border: 1px solid #dddddd;
              text-align: left;
              padding: 8px;
            }}

            tr:nth-child(even) {{
              background-color:
            }}
            </style>
            </head>
            <body>
            """
            )

            if preamble is not None:
                with open(preamble, "r") as f:
                    preamble = f.read()
                f.write(markdown.markdown(preamble))

            f.write(
                f"""
            <h1>Observing Summary</h1>
            <p>Number of observing projects: {n_codes}</p>
            <p>Number of images: {n_images}</p>
            <p>Total exposure time: {exp_tot:.2f} minutes</p>
            <p>Moon phase range: {100*moon_phs_range[0]:.2f}% -- {100*moon_phs_range[1]:.2f}%</p>
            <p>Median seeing: {seeing_median:.2f} +/- {seeing_std:.2f} arcsec</p>
            <p>Median pointing error: {dra_median:.2f} +/- {dra_std:.2f} arcsec (RA), {ddec_median:.2f} +/- {ddec_std:.2f} arcsec (Dec)</p>
            <p>Median late time: {late_median:.2f} +/- {late_std:.2f} minutes</p>
            """
            )

            f.write(
                f"""
            <h2>Summary Table</h2>
            {summary_table.pformat_all(html=True)}
            """
            )

            f.write(
                f"""
            <h2>Image Table</h2>
            {tbl.pformat_all(html=True)}
            """
            )

            f.write(
                f"""
            </body>
            </html>
            """
            )

        logger.info(f"Summary report written to {output}")

    if verbose == -1:
        return (
            n_codes,
            n_images,
            exp_tot,
            moon_phs_range,
            seeing_median,
            seeing_std,
            zmags,
            zmag_errs,
            dra_median,
            dra_std,
            ddec_median,
            ddec_std,
            late_median,
            late_std,
            summary_table,
            tbl,
        )


summary_report = summary_report_cli.callback



def _kwargs_to_config(kwargs):
    if kwargs is None or len(kwargs) == 0:
        return ""
    string = ""
    for key, value in kwargs.items():
        if ":" in str(value):
            string += str(key) + "='" + str(value) + "',"
        else:
            string += str(key) + "=" + str(value) + ","
    return string[:-1]



import logging
import os

import click
import matplotlib.pyplot as plt
import numpy as np
import tqdm
from astropy import units as u
from astropy import wcs
from astropy.io import fits
from astroquery import sdss

from ..observatory import AstrometryNetWCS
from ..utils import _get_image_source_catalog

logger = logging.getLogger(__name__)


@click.command(
    epilog="""Check out the documentation at
                https://pyscope.readthedocs.io/ for more
                information."""
)
@click.option(
    "-f",
    "--filter",
    "filt",
    type=click.Choice(["u", "b", "g", "r", "i", "z"]),
    show_choices=True,
    help="""Sloan filter to use for computing the
                zero-point magnitude. Overrides the header filter.""",
)
@click.option(
    "-b",
    "--background-params",
    "background_params",
    nargs=2,
    type=(float, float),
    default=(50, 3),
    show_default=True,
    help="""Background2D box_size and filter_size
                parameters in pixels.""",
)
@click.option(
    "-t",
    "--threshold",
    type=float,
    default=3.0,
    show_default=True,
    help="""Threshhold for source detection in units of
                background RMS.""",
)
@click.option(
    "-g",
    "--gaussian-params",
    nargs=2,
    type=(float, float),
    default=(5, 3),
    show_default=True,
    help="Gaussian2D sigma [pixels] and size parameters.",
)
@click.option(
    "-e",
    "--effective-gain",
    "effective_gain",
    type=float,
    default=1,
    show_default=True,
    help="""Effective gain for computing the Poisson noise
                contribution to the total error. In units of electrons
                or photons per count""",
)
@click.option(
    "-c",
    "--connectivity",
    type=click.Choice(["4", "8"]),
    default="8",
    show_default=True,
    show_choices=True,
    help="""Definition of pixel connectivity, i.e. whether
                diagonal pixels are considered to be connected (8 if
                yes, 4 if no).""",
)
@click.option(
    "-n",
    "--npixels",
    type=int,
    default=10,
    show_default=True,
    help="Minimum number of connected pixels",
)
@click.option(
    "-l",
    "--nlevels",
    type=int,
    default=32,
    show_default=True,
    help="""Number of multi-thresholding levels to use
                for source deblending.""",
)
@click.option(
    "-d",
    "--deblend-contrast",
    "deblend_contrast",
    type=float,
    default=0.005,
    show_default=True,
    help="""Contrast ratio used for source deblending.""",
)
@click.option(
    "-s",
    "--search-radius",
    "search_radius",
    type=float,
    default=3.0,
    help="""Radius to search for nearby
                sources in the SDSS catalog [arcsec].""",
)
@click.option(
    "-w",
    "--write",
    is_flag=True,
    default=False,
    help="""Write the ZMAG and ZMAGERR to the header.""",
)
@click.option("-p", "--plot", is_flag=True, default=False, help="""Plot the results.""")
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="""Print verbose output."""
)
@click.argument("images", type=click.Path(exists=True), nargs=-1)
@click.version_option()
def calc_zmag_cli(
    images,
    filt=None,
    background_params=(50, 3),
    threshold=25,
    gaussian_params=(5, 3),
    effective_gain=1,
    connectivity=8,
    npixels=10,
    nlevels=32,
    deblend_contrast=0.005,
    search_radius=3,
    write=False,
    plot=False,
    verbose=False,
):
    if verbose:
        logger.setLevel("DEBUG")

    logger.debug(
        f"""calc_zmag_cli(filt={filt}, background_params={background_params},
                    threshold={threshold}, gaussian_params={gaussian_params},
                    effective_gain={effective_gain}, connectivity={connectivity},
                    npixels={npixels}, levels={nlevels},
                    deblend_contrast={deblend_contrast}, search_radius={search_radius},
                    write={write}, plot={plot}, verbose={verbose})"""
    )
    logger.info("Starting calc_zmag")

    if type(images) == str:
        images = [images]

    zmags = []
    zmags_err = []

    for im in images:
        os.path.abspath(im)
        logger.info("Processing %s" % im)

        # Check if WCS solution exists already, if not, try one
        try:
            with fits.open(im) as hdul:
                header = hdul[0].header
            w = wcs.WCS(header)
            sky = w.pixel_to_world(0, 0)
        except:
            logger.info("No WCS solution found. Attempting to solve.")
            solver = AstrometryNetWCS()
            success = solver.Solve(im)
            if success:
                logger.info("Solved.")
            else:
                logger.error("Failed to solve, continuing to next image if present...")
                continue

        # Get the source catalog
        logger.info("Detecting sources...")
        catalog = _get_image_source_catalog(
            im,
            box_size=background_params[0],
            filter_size=background_params[1],
            threshold=threshold,
            kernel_fwhm=gaussian_params[0],
            kernel_size=gaussian_params[1],
            effective_gain=effective_gain,
            connectivity=connectivity,
            npixels=npixels,
            nlevels=nlevels,
            contrast=deblend_contrast,
        ).to_table(
            columns=(
                "sky_centroid_icrs",
                "xcentroid",
                "ycentroid",
                "segment_flux",
                "segment_fluxerr",
            )
        )

        logger.info("Found %d sources." % len(catalog))

        with fits.open(im) as hdul:
            data = hdul[0].data
            try:
                exp = hdul[0].header["EXPTIME"]
            except:
                exp = hdul[0].header["EXPOSURE"]
            im_filt = hdul[0].header["FILTER"]

        w = wcs.WCS(hdul[0].header)

        # Get header filter or override with CLI option
        if filt is None:
            filt = im_filt
        filt = filt.lower()
        logger.info("Filter: %s" % filt)

        if filt not in ["u", "b", "g", "r", "i", "z"]:
            logger.error("Invalid filter, continuing to next image...")
            continue

        catalog["SDSS"] = [None for i in range(len(catalog))]

        # SDSS lookup
        logger.info("Looking up corresponding SDSS sources...")
        for source in tqdm.tqdm(catalog):
            logger.debug("Source:\n%s" % source)

            potential_matches = sdss.SDSS.query_region(
                source["sky_centroid_icrs"],
                radius=search_radius * u.arcsec,
                cache=True,
                fields=["ra", "dec", "clean", filt],
            )

            try:
                for match in potential_matches:
                    if match["clean"] and match[filt] < 24:
                        source["SDSS"] = float(match[filt])
                        logger.debug("Match found:\n%s" % match)
                        break
            except:
                logger.debug("No SDSS matches found, continuing to next source...")
                continue

        catalog.remove_rows(catalog["SDSS"] == None)

        logger.info("Found %d SDSS matches." % len(catalog))

        # Compute zero-point magnitude
        logger.info("Computing zero-point magnitude...")

        mag = catalog["SDSS"].data
        flux = catalog["segment_flux"].data / exp
        flux_err = catalog["segment_fluxerr"].data / exp

        zmag = mag + 2.5 * np.log10(flux)
        zmag_err = 2.5 * np.log10(1 + flux_err / flux)

        clip_array = (zmag < np.mean(zmag) + 2 * np.std(zmag)) & (
            zmag > np.mean(zmag) - 2 * np.std(zmag)
        )
        zmag = zmag[clip_array]
        zmag_err = zmag_err[clip_array]
        mag = mag[clip_array]
        flux = flux[clip_array]
        flux_err = flux_err[clip_array]

        zmags.append(zmag)
        zmags_err.append(zmag_err)

        mean_zmag = np.mean(zmag)
        mean_zmag_err = np.sqrt(np.sum(zmag_err**2)) / len(zmag)

        logger.info(
            "Mean zero-point magnitude: %.3f +/- %.3f" % (mean_zmag, mean_zmag_err)
        )

        if write:
            logger.info("Writing to header...")
            hdul[0].header["ZMAG"] = mean_zmag
            hdul[0].header["ZMAGERR"] = mean_zmag_err
            hdul.writeto(im, overwrite=True)

        if plot:
            fig = plt.figure(figsize=(16, 16), dpi=400)
            ax0 = fig.add_subplot(3, 2, (1, 4), projection=w)
            ax1 = fig.add_subplot(3, 2, 5)
            ax2 = fig.add_subplot(3, 2, 6)
            fig.subplots_adjust(wspace=0.15, hspace=0.1)

            # Left plot shows image with sources circled
            ax0.imshow(
                data,
                origin="lower",
                cmap="gray",
                vmin=np.percentile(data, 5),
                vmax=np.percentile(data, 99),
            )
            overlay = ax0.get_coords_overlay("icrs")
            overlay.grid(color="white", lw=2)

            for source in catalog:
                ax0.scatter(
                    source["xcentroid"], source["ycentroid"], marker="o", s=5, color="r"
                )
                ax0.text(
                    source["xcentroid"],
                    source["ycentroid"],
                    ("%.2f" % source["SDSS"]),
                    color="r",
                    fontsize=12,
                )

            ax0.set_xlabel(" ")
            ax0.set_ylabel(" ")

            # Right plot shows flux vs. magnitude
            ax1.errorbar(
                mag, flux, yerr=flux_err, fmt="", linestyle="none", color="black"
            )
            for i in range(len(mag)):
                ax1.text(
                    mag[i], flux[i], ("%.1f" % flux[i]), color="black", fontsize=12
                )

            model_mag = np.linspace(14, 24, 100)
            model_flux = 10 ** (-0.4 * (model_mag - mean_zmag))
            ax1.plot(model_mag, model_flux, color="black")

            ax1.axvline(x=mean_zmag, color="red", ls="dashed")
            ax1.text(
                mean_zmag - 0.05,
                0.85,
                ("%.2f +/- %.3f" % (mean_zmag, mean_zmag_err)),
                color="red",
                fontsize=12,
            )

            ax1.set_xlim(24, 14)
            ax1.set_yscale("log")

            ax1.set_xlabel("SDSS Magnitude")
            ax1.set_ylabel("Flux [counts/sec]")

            ax2.scatter(mag, flux / flux_err, color="black", marker=".", s=5)
            for i in range(len(mag)):
                ax2.text(
                    mag[i],
                    flux[i] / flux_err[i],
                    ("%.2f" % (flux[i] / flux_err[i])),
                    color="black",
                    fontsize=12,
                )

            ax2.set_xlim(24, 14)
            ax2.set_ylim(threshold)
            ax2.set_yscale("log")
            ax2.set_xlabel("SDSS Magnitude")
            ax2.set_ylabel("SNR")

            fig.suptitle(
                "%s\nZMAG: %.2f +/- %.3f" % (im, mean_zmag, mean_zmag_err),
                y=0.95,
                fontsize=16,
            )

            # plt.show()

    logger.info("Done.")

    return zmags, zmags_err, fig, (ax0, ax1, ax2)


calc_zmag = calc_zmag_cli.callback



import binascii
import logging
import socket
import struct
import sys

from .cover_calibrator import CoverCalibrator

logger = logging.getLogger(__name__)


class IPCoverCalibrator(CoverCalibrator):
    def __init__(self, tcp_ip, tcp_port, buffer_size):
        self._tcp_ip = tcp_ip
        self._tcp_port = tcp_port
        self._buffer_size = buffer_size

    def CalibratorOff(self):
        logger.debug("IPCoverCalibrator.CalibratorOff called")
        return self._send_packet(0)

    def CalibratorOn(self, Brightness):
        logger.debug(f"CalibratorOn called with Brightness={Brightness}")
        return self._send_packet(Brightness)

    def CloseCover(self):
        logger.debug("IPCoverCalibrator.CloseCover called")
        raise NotImplementedError

    def HaltCover(self):
        logger.debug("IPCoverCalibrator.HaltCover called")
        raise NotImplementedError

    def OpenCover(self):
        logger.debug("IPCoverCalibrator.OpenCover called")
        raise NotImplementedError

    @property
    def Brightness(self):
        logger.debug("IPCoverCalibrator.Brightness called")
        return

    @property
    def CalibratorState(self):
        logger.debug("IPCoverCalibrator.CalibratorState called")
        return

    @property
    def CoverState(self):
        logger.debug("IPCoverCalibrator.CoverState called")
        return

    @property
    def MaxBrightness(self):
        logger.debug("IPCoverCalibrator.MaxBrightness called")
        return 254

    def _send_packet(self, intensity):
        logger.debug(f"_send_packet called with intensity={intensity}")
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.tcp_ip, self.tcp_port))

        # Create 4 byte array
        my_bytes = bytearray()
        my_bytes.append(253)
        my_bytes.append(0)
        my_bytes.append(intensity)
        my_bytes.append(1)

        try:
            s.send(my_bytes)
        except:
            return False
        finally:
            data = s.recv(self.buffer_size)
            if data != "U":
                return False
            s.close()
            return True

    @property
    def tcp_ip(self):
        logger.debug("IPCoverCalibrator.tcp_ip called")
        return self._tcp_ip

    @property
    def tcp_port(self):
        logger.debug("IPCoverCalibrator.tcp_port called")
        return self._tcp_port

    @property
    def buffer_size(self):
        logger.debug("IPCoverCalibrator.buffer_size called")
        return self._buffer_size



# Built-in Python imports
import logging
import queue

# iotalib imports
from . import config_notification, gmail, logutil


def send_info(subject_suffix, message):
    if not config_notification.valid_config:
        logging.error(
            "Not sending e-mail notification: don't have a valid notification config"
        )
        return

    subject = "INFO - " + subject_suffix

    send_mail(config_notification.values.info_emails, subject, message)


def send_warnings():
    if not config_notification.valid_config:
        logging.error(
            "Not sending e-mail notification: don't have a valid notification config"
        )
        return

    warnings = []
    while True:
        try:
            warning_message = logutil.warning_queue.get_nowait()
            warnings.append(warning_message)
        except queue.Empty:
            break

    subject = "%d Recent IOTA Warnings" % len(warnings)
    message = "\n".join(warnings)

    send_mail(config_notification.values.error_emails, subject, message)


def send_error(subject_suffix, message):
    if not config_notification.valid_config:
        logging.error(
            "Not sending e-mail notification: don't have a valid notification config"
        )
        return

    subject = "ERROR - " + subject_suffix

    try:
        log_messages = []
        while True:
            try:
                log_message = logutil.recent_message_queue.get_nowait()
                log_messages.append(log_message)
            except queue.Empty:
                break
        log_messages_text = "%d most recent log messages:\n%s" % (
            len(log_messages),
            "\n".join(log_messages),
        )

        message = message + "\n\n" + log_messages_text
    except:
        pass  # Don't let a problem with our log message retrieval prevent us from getting the message out

    send_mail(config_notification.values.error_emails, subject, message)


def send_mail(email_addresses, subject, message):
    if config_notification.values.simulate_email:
        logging.info(
            "SIMULATED EMAIL: Recipients = '%r', Subject = '%s', Message = '%s'",
            email_addresses,
            subject,
            message,
        )
    else:
        gmail.send_mail(
            config_notification.values.gmail_username,
            config_notification.values.gmail_password,
            email_addresses,
            subject,
            message,
        )



import os

import click

from .syncfiles import syncfiles
from .telrun_operator import TelrunOperator


@click.command()
@click.argument(
    "path", type=click.Path(resolve_path=True), default="./", required=False
)
@click.option(
    "-g", "--gui", is_flag=True, default=True, show_default=True, help="Start the GUI"
)
@click.version_option()
def start_telrun_cli(path, gui):
    telrun = TelrunOperator(
        config_path=os.path.join(path, "config/telrun.cfg"), gui=gui
    )
    telrun.mainloop()


@click.command()
@click.argument(
    "path", type=click.Path(resolve_path=True), default="./", required=False
)
@click.version_option()
def start_syncfiles_cli(path):
    syncfiles(config=os.path.join(path, "config/"))


start_telrun = start_telrun_cli.callback
start_syncfiles = start_syncfiles_cli.callback
